#
# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from copy import deepcopy
import os.path

import pytest

from fontTools import designspaceLib
from glyphsLib import to_glyphs, to_designspace, to_ufos
from glyphsLib.classes import GSFont, GSFontMaster, GSAxis, GSInstance
from glyphsLib.builder.axes import _is_subset_of_default_axes, get_regular_master

"""
Goal: check how files with custom axes are roundtripped.
"""


@pytest.mark.parametrize(
    "axes",
    [
        [("wght", "Weight alone")],
        [("wdth", "Width alone")],
        [("XXXX", "Custom alone")],
        [("wght", "Weight (with width)"), ("wdth", "Width (with weight)")],
        [
            ("wght", "Weight (1/3 default)"),
            ("wdth", "Width (2/3 default)"),
            ("XXXX", "Custom (3/3 default)"),
        ],
        [("ABCD", "First custom"), ("EFGH", "Second custom")],
        [
            ("ABCD", "First custom"),
            ("EFGH", "Second custom"),
            ("IJKL", "Third custom"),
            ("MNOP", "Fourth custom"),
        ],
        [("opsz", "First custom"), ("wght", "Second custom")],
        # Test that standard axis definitions don't generate an Axes custom parameter.
        [("wght", "Weight"), ("wdth", "Width")],
        [("wdth", "Width"), ("wght", "Weight")],
    ],
)
def test_weight_width_custom(axes, ufo_module):
    """Test that having axes in any order or quantity does not confuse
    glyphsLib, even when the weight or width are not in the default positions.
    """
    doc = _make_designspace_with_axes(axes, ufo_module)

    font = to_glyphs(doc)

    if _is_subset_of_default_axes([GSAxis(name=n, tag=t) for t, n in axes]):
        assert font.customParameters["Axes"] is None
    else:
        assert font.customParameters["Axes"] == [
            {"Tag": tag, "Name": name} for tag, name in axes
        ]

    doc = to_designspace(font, ufo_module=ufo_module)

    assert len(doc.axes) == len(axes)
    for doc_axis, (tag, name) in zip(doc.axes, axes):
        assert doc_axis.tag == tag
        assert doc_axis.name == name


def _make_designspace_with_axes(axes, ufo_module):
    doc = designspaceLib.DesignSpaceDocument()

    # Add a "Regular" source
    regular = doc.newSourceDescriptor()
    regular.font = ufo_module.Font()
    regular.location = {name: 0 for _, name in axes}
    doc.addSource(regular)

    for tag, name in axes:
        axis = doc.newAxisDescriptor()
        axis.tag = tag
        axis.name = name
        axis.minimum = 0
        axis.default = 0
        axis.maximum = 100
        doc.addAxis(axis)

        extreme = doc.newSourceDescriptor()
        extreme.font = ufo_module.Font()
        extreme.location = {name_: 0 if name_ != name else 100 for _, name_ in axes}
        doc.addSource(extreme)

    return doc


def test_masters_have_user_locations(ufo_module):
    """Test the new axis definition with custom parameters.
    See https://github.com/googlefonts/glyphsLib/issues/280.

    For tests about the previous system with weight/width/custom,
    see `tests/builder/interpolation_test.py`.
    """
    # Get a font with two masters
    font = to_glyphs([ufo_module.Font(), ufo_module.Font()])
    font.customParameters["Axes"] = [{"Tag": "opsz", "Name": "Optical"}]
    # There is only one axis, so the design location is stored in the weight
    font.masters[0].weightValue = 0
    # The user location is stored as a custom parameter
    font.masters[0].customParameters["Axis Location"] = [
        {"Axis": "Optical", "Location": 13}
    ]
    font.masters[1].weightValue = 1000
    font.masters[1].customParameters["Axis Location"] = [
        {"Axis": "Optical", "Location": 100}
    ]

    doc = to_designspace(font, ufo_module=ufo_module)
    assert len(doc.axes) == 1
    assert doc.axes[0].map == [(13, 0), (100, 1000)]
    assert len(doc.sources) == 2
    assert doc.sources[0].location == {"Optical": 0}
    assert doc.sources[1].location == {"Optical": 1000}

    font = to_glyphs(doc)
    assert font.customParameters["Axes"] == [{"Tag": "opsz", "Name": "Optical"}]
    assert font.masters[0].weightValue == 0
    assert font.masters[0].customParameters["Axis Location"] == [
        {"Axis": "Optical", "Location": 13}
    ]
    assert font.masters[1].weightValue == 1000
    assert font.masters[1].customParameters["Axis Location"] == [
        {"Axis": "Optical", "Location": 100}
    ]


def test_masters_have_user_locations_string(ufo_module):
    """Test that stringified Axis Locations are converted.

    Some versions of Glyph store a string instead of an int.
    """
    font = to_glyphs([ufo_module.Font(), ufo_module.Font()])
    font.customParameters["Axes"] = [{"Tag": "opsz", "Name": "Optical"}]
    font.masters[0].weightValue = 0
    font.masters[0].customParameters["Axis Location"] = [
        {"Axis": "Optical", "Location": 13}
    ]
    font.masters[1].weightValue = 1000
    font.masters[1].customParameters["Axis Location"] = [
        {"Axis": "Optical", "Location": "100"}
    ]

    doc = to_designspace(font, ufo_module=ufo_module)
    assert doc.axes[0].map == [(13, 0), (100, 1000)]

    font = to_glyphs(doc)
    assert font.masters[0].customParameters["Axis Location"] == [
        {"Axis": "Optical", "Location": 13}
    ]
    assert font.masters[1].customParameters["Axis Location"] == [
        {"Axis": "Optical", "Location": 100}
    ]


def test_master_user_location_goes_into_os2_classes(ufo_module):
    font = to_glyphs([ufo_module.Font(), ufo_module.Font()])
    font.customParameters["Axes"] = [
        {"Tag": "wght", "Name": "Weight"},
        {"Tag": "wdth", "Name": "Width"},
    ]
    font.masters[0].weightValue = 0
    font.masters[0].widthValue = 1000
    # This master will be Light Expanded
    # as per https://docs.microsoft.com/en-gb/typography/opentype/spec/os2#uswidthclass
    font.masters[0].customParameters["Axis Location"] = [
        {"Axis": "Weight", "Location": 300},
        {"Axis": "Width", "Location": 125},
    ]
    font.masters[1].weightValue = 1000
    font.masters[1].widthValue = 0
    # This master is Black Ultra-condensed but not quite
    font.masters[1].customParameters["Axis Location"] = [
        {"Axis": "Weight", "Location": 920},  # instead of 900
        {"Axis": "Width", "Location": 55},  # instead of 50
    ]

    light, black = to_ufos(font)

    assert light.info.openTypeOS2WeightClass == 300
    assert light.info.openTypeOS2WidthClass == 7

    assert black.info.openTypeOS2WeightClass == 920
    assert black.info.openTypeOS2WidthClass == 1


def test_mapping_is_same_regardless_of_axes_custom_parameter(ufo_module):
    # https://github.com/googlefonts/glyphsLib/issues/409
    # https://github.com/googlefonts/glyphsLib/issues/411

    # First, try without the custom param
    font = to_glyphs([ufo_module.Font(), ufo_module.Font(), ufo_module.Font()])
    font.masters[0].name = "ExtraLight"
    font.masters[0].weightValue = 200
    font.masters[1].name = "Regular"
    font.masters[1].weightValue = 400
    font.masters[2].name = "Bold"
    font.masters[2].weightValue = 700

    doc = to_designspace(font, ufo_module=ufo_module)
    assert doc.axes[0].minimum == 200
    assert doc.axes[0].maximum == 700
    assert doc.axes[0].map == []

    # Now with the custom parameter. Should produce the same results
    font.customParameters["Axes"] = [{"Name": "Weight", "Tag": "wght"}]

    doc = to_designspace(font, ufo_module=ufo_module)
    assert doc.axes[0].minimum == 200
    assert doc.axes[0].maximum == 700
    assert doc.axes[0].map == []


def test_mapping_using_axis_location_custom_parameter_on_instances(ufo_module):
    # https://github.com/googlefonts/glyphsLib/issues/714
    # https://github.com/googlefonts/glyphsLib/pull/810

    font = to_glyphs(
        [ufo_module.Font(), ufo_module.Font(), ufo_module.Font(), ufo_module.Font()]
    )

    origin_id = "95FB0C11-C828-4064-8966-34220AA4D426"
    font.customParameters["Axes"] = [
        {"Name": "Weight", "Tag": "wght"},
        {"Name": "Width", "Tag": "wdth"},
    ]

    # Add masters

    font.masters[0].name = "Regular"
    font.masters[0].id = origin_id
    font.masters[0].weightValue = 72
    font.masters[0].widthValue = 448
    font.masters[0].customParameters["Axis Location"] = [
        {"Axis": "Weight", "Location": 400},
        {"Axis": "Width", "Location": 100},
    ]
    font.masters[1].name = "Bold"
    font.masters[1].weightValue = 112
    font.masters[1].widthValue = 448
    font.masters[1].customParameters["Axis Location"] = [
        {"Axis": "Weight", "Location": 700},
        {"Axis": "Width", "Location": 100},
    ]
    font.masters[2].name = "Thin"
    font.masters[2].weightValue = 48
    font.masters[2].widthValue = 448
    font.masters[2].customParameters["Axis Location"] = [
        {"Axis": "Weight", "Location": 200},
        {"Axis": "Width", "Location": 100},
    ]
    font.masters[3].name = "Cd Regular"
    font.masters[3].weightValue = 72
    font.masters[3].widthValue = 224
    font.masters[3].customParameters["Axis Location"] = [
        {"Axis": "Weight", "Location": 400},
        {"Axis": "Width", "Location": 50},
    ]

    font.customParameters["Variable Font Origin"] = origin_id

    # Add some instances with mappings

    font.instances = [
        GSInstance(),
        GSInstance(),
        GSInstance(),
        GSInstance(),
        GSInstance(),
        GSInstance(),
    ]

    font.instances[0].name = "Thin"
    font.instances[0].weightValue = 48
    font.instances[0].widthValue = 448
    font.instances[0].customParameters["Axis Location"] = [
        {"Axis": "Weight", "Location": 200},
        {"Axis": "Width", "Location": 100},
    ]
    font.instances[1].name = "Light"
    font.instances[1].weightValue = 62
    font.instances[1].widthValue = 448
    font.instances[1].customParameters["Axis Location"] = [
        {"Axis": "Weight", "Location": 300},
        {"Axis": "Width", "Location": 100},
    ]
    font.instances[2].name = "Regular"
    font.instances[2].weightValue = 72
    font.instances[2].widthValue = 448
    font.instances[2].customParameters["Axis Location"] = [
        {"Axis": "Weight", "Location": 400},
        {"Axis": "Width", "Location": 100},
    ]
    font.instances[3].name = "Medium"
    font.instances[3].weightValue = 92
    font.instances[3].widthValue = 448
    font.instances[3].customParameters["Axis Location"] = [
        {"Axis": "Weight", "Location": 600},
        {"Axis": "Width", "Location": 100},
    ]
    font.instances[4].name = "Cd Regular"
    font.instances[4].weightValue = 72
    font.instances[4].widthValue = 224
    font.instances[4].customParameters["Axis Location"] = [
        {"Axis": "Weight", "Location": 400},
        {"Axis": "Width", "Location": 50},
    ]
    font.instances[5].name = "SCd Regular"
    font.instances[5].weightValue = 72
    font.instances[5].widthValue = 384
    font.instances[5].customParameters["Axis Location"] = [
        {"Axis": "Weight", "Location": 400},
        {"Axis": "Width", "Location": 60},
    ]

    doc = to_designspace(font, ufo_module=ufo_module)
    assert doc.axes[0].minimum == 200
    assert doc.axes[0].default == 400
    assert doc.axes[0].maximum == 700
    assert doc.axes[0].map == [(200, 48), (300, 62), (400, 72), (600, 92), (700, 112)]

    assert doc.axes[1].minimum == 50
    assert doc.axes[1].default == 100
    assert doc.axes[1].maximum == 100
    assert doc.axes[1].map == [(50, 224), (60, 384), (100, 448)]


def test_mapping_using_axis_location_cp_on_masters_none(ufo_module):
    # https://github.com/googlefonts/glyphsLib/issues/714
    # https://github.com/googlefonts/glyphsLib/pull/810

    # When masters have no or disabled Axis Location CP, the ones on the
    # instances should still be evaluated.

    font = to_glyphs(
        [ufo_module.Font(), ufo_module.Font(), ufo_module.Font(), ufo_module.Font()]
    )

    font.customParameters["Axes"] = [
        {"Name": "Weight", "Tag": "wght"},
        {"Name": "Width", "Tag": "wdth"},
    ]

    # Add masters

    font.masters[0].name = "Regular"
    font.masters[0].weightValue = 72
    font.masters[0].widthValue = 448

    font.masters[1].name = "Bold"
    font.masters[1].weightValue = 112
    font.masters[1].widthValue = 448

    font.masters[2].name = "Thin"
    font.masters[2].weightValue = 48
    font.masters[2].widthValue = 448

    font.masters[3].name = "Cd Regular"
    font.masters[3].weightValue = 72
    font.masters[3].widthValue = 224

    font.instances = [GSInstance(), GSInstance(), GSInstance()]

    font.instances[0].name = "Regular"
    font.instances[0].weightValue = 72
    font.instances[0].widthValue = 448
    font.instances[0].customParameters["Axis Location"] = [
        {"Axis": "Weight", "Location": 400},
        {"Axis": "Width", "Location": 100},
    ]
    font.instances[1].name = "SCd Regular"
    font.instances[1].weightValue = 72
    font.instances[1].widthValue = 384
    font.instances[1].customParameters["Axis Location"] = [
        {"Axis": "Weight", "Location": 400},
        {"Axis": "Width", "Location": 60},
    ]
    font.instances[2].name = "Cd Regular"
    font.instances[2].weightValue = 72
    font.instances[2].widthValue = 224
    font.instances[2].customParameters["Axis Location"] = [
        {"Axis": "Weight", "Location": 400},
        {"Axis": "Width", "Location": 50},
    ]

    doc = to_designspace(font, ufo_module=ufo_module)
    assert doc.axes[0].minimum == 400
    assert doc.axes[0].default == 400
    assert doc.axes[0].maximum == 400
    assert doc.axes[0].map == [(400, 72)]

    assert doc.axes[1].minimum == 50
    assert doc.axes[1].default == 100
    assert doc.axes[1].maximum == 100
    assert doc.axes[1].map == [(50, 224), (60, 384), (100, 448)]


def test_mapping_using_axis_location_cp_on_instances_none(ufo_module):
    # https://github.com/googlefonts/glyphsLib/issues/714
    # https://github.com/googlefonts/glyphsLib/pull/810

    # When all masters have Axis Location CP, non-"Axis Location" instance
    # mappings should be ignored.

    font = to_glyphs(
        [ufo_module.Font(), ufo_module.Font(), ufo_module.Font(), ufo_module.Font()]
    )

    font.customParameters["Axes"] = [
        {"Name": "Weight", "Tag": "wght"},
        {"Name": "Width", "Tag": "wdth"},
    ]

    # Add masters

    font.masters[0].name = "Regular"
    font.masters[0].weightValue = 72
    font.masters[0].widthValue = 448
    font.masters[0].customParameters["Axis Location"] = [
        {"Axis": "Weight", "Location": 400},
        {"Axis": "Width", "Location": 100},
    ]
    font.masters[1].name = "Bold"
    font.masters[1].weightValue = 112
    font.masters[1].widthValue = 448
    font.masters[1].customParameters["Axis Location"] = [
        {"Axis": "Weight", "Location": 700},
        {"Axis": "Width", "Location": 100},
    ]
    font.masters[2].name = "Thin"
    font.masters[2].weightValue = 48
    font.masters[2].widthValue = 448
    font.masters[2].customParameters["Axis Location"] = [
        {"Axis": "Weight", "Location": 200},
        {"Axis": "Width", "Location": 100},
    ]
    font.masters[3].name = "Cd Regular"
    font.masters[3].weightValue = 72
    font.masters[3].widthValue = 224
    font.masters[3].customParameters["Axis Location"] = [
        {"Axis": "Weight", "Location": 400},
        {"Axis": "Width", "Location": 50},
    ]

    font.instances = [GSInstance()]

    font.instances[0].name = "SCd Regular"
    font.instances[0].weightValue = 72
    font.instances[0].widthValue = 384

    doc = to_designspace(font, ufo_module=ufo_module)
    assert doc.axes[0].minimum == 200
    assert doc.axes[0].default == 400
    assert doc.axes[0].maximum == 700
    assert doc.axes[0].map == [(200, 48), (400, 72), (700, 112)]

    assert doc.axes[1].minimum == 50
    assert doc.axes[1].default == 100
    assert doc.axes[1].maximum == 100
    assert doc.axes[1].map == [(50, 224), (100, 448)]


def test_custom_parameter_vfo_current():
    """Tests get_regular_master when 'Variable Font Origin' custom parameter name
    is used with master set to 'Regular Text'.  This is the current default
    custom parameter name in the Glyphs editor / glyphs source file specification."""
    source_path = os.path.join("tests", "data", "CustomParameterVFO.glyphs")
    font = GSFont(source_path)
    assert font.customParameters["Variation Font Origin"] is None
    test_id = font.customParameters["Variable Font Origin"]
    assert test_id == "ACC63F3E-1323-486A-94AF-B18797A154CE"
    matched = False
    for master in font.masters:
        if master.id == test_id:
            assert master.name == "Regular Text"
            matched = True
    assert matched is True
    default_master = get_regular_master(font)
    assert default_master.name == "Regular Text"


def test_custom_parameter_vfo_old_name():
    """Tests get_regular_master when 'Variation Font Origin' custom parameter name
    is used with master set to 'Regular Text'.  This custom parameter name is not
    used in current releases of the Glyphs editor / glyphs source file specification."""
    source_path = os.path.join("tests", "data", "CustomParameterVFO.glyphs")
    font = GSFont(source_path)

    # mock up source for this test with a source file from another test
    del font.customParameters["Variable Font Origin"]
    font.customParameters["Variation Font Origin"] = "Regular Text"
    assert font.customParameters["Variable Font Origin"] is None
    # start tests
    test_name = font.customParameters["Variation Font Origin"]
    matched = False
    for master in font.masters:
        if master.name == test_name:
            matched = True
    assert matched is True
    default_master = get_regular_master(font)
    assert default_master.name == "Regular Text"


def test_custom_parameter_vfo_not_set():
    """Tests default behavior of get_regular_master when Variable Font Origin custom
    parameter is not set"""
    source_path = os.path.join("tests", "data", "CustomParameterVFO.glyphs")
    font = GSFont(source_path)

    # mock up source for this test with a source file from another test
    del font.customParameters["Variable Font Origin"]
    del font.customParameters["Variation Font Origin"]
    assert font.customParameters["Variable Font Origin"] is None
    assert font.customParameters["Variation Font Origin"] is None
    default_master = get_regular_master(font)
    assert default_master.name == "Regular Text"


def test_wheres_ma_axis(datadir):
    font1 = GSFont(datadir.join("AxesWdth.glyphs"))
    doc1 = to_designspace(font1)
    assert [a.tag for a in doc1.axes] == ["wdth"]


def test_wheres_ma_axis2(datadir):
    font2 = GSFont(datadir.join("AxesWdthWght.glyphs"))
    doc2 = to_designspace(font2)
    assert [a.tag for a in doc2.axes] == ["wdth", "wght"]


def test_single_master_default_weight_400(ufo_module):
    font = GSFont()
    master = GSFontMaster()
    master.weightValue = 400
    font.masters.append(master)

    doc = to_designspace(font, ufo_module=ufo_module)

    assert len(doc.axes) == 1
    assert doc.axes[0].name == "Weight"
    assert doc.axes[0].minimum == 400
    assert doc.axes[0].default == 400
    assert doc.axes[0].maximum == 400
    assert len(doc.sources) == 1
    assert doc.sources[0].location["Weight"] == 400

    font2 = to_glyphs(doc)

    assert len(font2.masters) == 1
    assert font2.masters[0].weightValue == 400


def test_axis_mapping(ufo_module):
    font = to_glyphs(
        [ufo_module.Font(), ufo_module.Font(), ufo_module.Font(), ufo_module.Font()]
    )
    font.masters[0].weightValue = 0
    font.masters[0].widthValue = 100
    font.masters[1].weightValue = 1000
    font.masters[1].widthValue = 100

    font.masters[2].weightValue = 0
    font.masters[2].widthValue = 75
    font.masters[3].weightValue = 1000
    font.masters[3].widthValue = 75

    wght_mapping = [(100, 0), (400, 350), (900, 1000)]
    wdth_mapping = [(75, 75), (100, 100)]

    axis_mappings = {
        "wght": {str(float(k)): v for k, v in wght_mapping},
        "wdth": {str(float(k)): v for k, v in wdth_mapping},
    }

    font.customParameters["Axis Mappings"] = axis_mappings
    # When we convert to a designspace, the wdth mapping is removed because
    # it isn't needed.
    doc = to_designspace(font, ufo_module=ufo_module)

    assert doc.axes[0].name == "Weight"
    assert doc.axes[0].minimum == 100
    assert doc.axes[0].default == 100
    assert doc.axes[0].maximum == 900
    assert doc.axes[0].map == wght_mapping

    assert doc.axes[1].name == "Width"
    assert doc.axes[1].minimum == 75
    assert doc.axes[0].default == 100
    assert doc.axes[1].maximum == 100
    assert doc.axes[1].map != wdth_mapping
    assert doc.axes[1].map == []

    font = to_glyphs(doc)
    assert font.customParameters["Axis Mappings"] == axis_mappings


def test_axis_with_no_mapping_does_not_error_in_roundtrip(ufo_module):
    """Tests that a custom axis without a mapping and without sources on its
    extremes does not generate an error during roundtrip. Also tests that
    during a to_glyphs, to_designspace roundtrip the min and max axis
    information is not lost.
    """
    doc = designspaceLib.DesignSpaceDocument()

    # Add a "Regular" source
    regular = doc.newSourceDescriptor()
    regular.font = ufo_module.Font()
    regular.location = {"Style": 0}
    doc.addSource(regular)

    axis = doc.newAxisDescriptor()
    axis.tag = "styl"
    axis.name = "Style"
    doc.addAxis(axis)

    # This axis spans a range of 0 to 1 but only has a source at {"Style": 0}
    # and no explicit mapping. The point of this test is to see if the min and
    # max are still the same after round tripping.
    doc.axes[0].minimum = 0
    doc.axes[0].maximum = 1
    doc.axes[0].default = 0
    doc.axes[0].map = []

    doc2 = deepcopy(doc)
    font = to_glyphs(doc2)
    doc_rt = to_designspace(font)

    assert doc_rt.axes[0].serialize() == doc.axes[0].serialize()


def test_axis_with_no_mapping_does_not_error_in_roundtrip_with_2_axes(ufo_module):
    """Tests that a designspace with 2 axis, one with a mapping and one
    without a mapping, roundtrips correctly without error. The axis without a
    mapping should generate an identity mapping on the fly so that the
    Glyphs.app customParameter field does not lose min/max infromation about
    the axis.
    """
    doc = _make_designspace_with_axes(
        [("wght", "Weight with mapping"), ("wdth", "Width without mapping")], ufo_module
    )
    # Add mapping to weight axis
    doc.axes[0].map = [(0, 0), (50, 350), (100, 1000)]

    doc2 = deepcopy(doc)
    font = to_glyphs(doc2)
    doc_rt = to_designspace(font)

    assert doc_rt.axes[0].serialize() == doc.axes[0].serialize()
    assert doc_rt.axes[1].serialize() == doc.axes[1].serialize()


def test_variable_instance(ufo_module):
    """Glyphs 3 introduced a "variable" instance which is a special instance
    that holds various VF settings. We export it to Design Space variable-font.
    """
    source_path = os.path.join("tests", "data", "VariableInstance.glyphs")
    font = GSFont(source_path)
    assert len(font.instances) == 28  # Including the VF setting
    doc = to_designspace(font)

    assert doc.axes[0].map[2] == (400, 80)
    assert doc.axes[0].default == 400
    assert len(doc.instances) == 27  # The VF setting should not be in the DS

    assert len(doc.variableFonts) == 1

    varfont = doc.variableFonts[0]
    assert varfont.name == "Variable Foo Bar"
    assert varfont.filename == "Cairo-VariableFooBarVF"
    assert len(varfont.axisSubsets) == len(doc.axes)
    assert "public.fontInfo" in varfont.lib

    info = varfont.lib["public.fontInfo"]
    assert len(info.get("openTypeNameRecords")) == 1
    assert info["openTypeNameRecords"][0].string == "Variable"
    assert info["openTypeNameRecords"][0].nameID == 1
    assert info["openTypeNameRecords"][0].platformID == 1
    assert info["openTypeNameRecords"][0].languageID == 0
    assert info["openTypeNameRecords"][0].encodingID == 0
    assert info.get("openTypeNamePreferredFamilyName") == "Cairo Variable"


def test_virtual_masters_extend_min_max_for_unmapped_axis(ufo_module, datadir):
    # https://github.com/googlefonts/glyphsLib/issues/859
    font = GSFont(datadir.join("IntermediateLayer.glyphs"))
    assert ["Cap Height", "Weight"] == [a.name for a in font.axes]

    assert "Axis Mappings" not in font.customParameters
    for master in font.masters:
        assert "Axis Location" not in master.customParameters
        # all non-virtual masters are at the default Cap Height location
        assert master.axes[0] == 700

    virtual_masters = [
        cp.value for cp in font.customParameters if cp.name == "Virtual Master"
    ]
    assert virtual_masters[0] == [
        {"Axis": "Cap Height", "Location": 600},
        {"Axis": "Weight", "Location": 400},
    ]
    assert virtual_masters[1] == [
        {"Axis": "Cap Height", "Location": 800},
        {"Axis": "Weight", "Location": 400},
    ]

    ds = to_designspace(font, ufo_module=ufo_module)

    # the min/max for this axis are taken from the virtual masters
    assert ds.axes[0].name == "Cap Height"
    assert ds.axes[0].minimum == 600
    assert ds.axes[0].default == 700
    assert ds.axes[0].maximum == 800
    assert not ds.axes[0].map

    assert ds.axes[1].name == "Weight"
    assert ds.axes[1].minimum == 400
    assert ds.axes[1].default == 400
    assert ds.axes[1].maximum == 900
    assert not ds.axes[1].map

    font2 = to_glyphs(ds)

    assert [
        cp.value for cp in font2.customParameters if cp.name == "Virtual Master"
    ] == virtual_masters


def test_axis_location_cp_uses_floats(ufo_module, datadir):
    # https://github.com/googlefonts/glyphsLib/issues/1100
    font = GSFont(datadir.join("WghtVar_AxisLocationFloat.glyphs"))

    assert len(font.masters) == 1
    assert font.masters[0].customParameters["Axis Location"] == [
        {
            "Axis": "Weight",
            "Location": 400.75,
        }
    ]

    ds = to_designspace(font, ufo_module=ufo_module)

    assert len(ds.sources) == 1
    assert ds.axes[0].name == "Weight"
    assert ds.axes[0].map == [(400.75, 0)]


def test_hidden_axis(ufo_module):
    font = to_glyphs([ufo_module.Font(), ufo_module.Font()])
    axes = [
        {"Tag": "opsz", "Name": "Optical"},
        {"Tag": "TEST", "Name": "Test Axis", "Hidden": 1},
    ]
    font.customParameters["Axes"] = axes
    doc = to_designspace(font, ufo_module=ufo_module)
    assert len(doc.axes) == 2
    assert not doc.axes[0].hidden
    assert doc.axes[1].hidden

    font = to_glyphs(doc)
    assert font.customParameters["Axes"] == axes



from typing import Any, Dict, Tuple, Union, Optional

from fontTools.pens.pointPen import AbstractPointPen

from glyphsLib.types import Transform, Point

Number = Union[int, float]


class LayerPointPen(AbstractPointPen):
    """A point pen to draw onto GSLayer object.

    See :mod:`fontTools.pens.basePen` and :mod:`fontTools.pens.pointPen` for an
    introduction to pens.
    """

    def __init__(self, layer: "GSLayer") -> None:  # noqa: F821
        self._layer: "GSLayer" = layer  # noqa: F821
        self._path: Optional["GSPath"] = None  # noqa: F821

    def beginPath(self, **kwargs: Any) -> None:
        from glyphsLib.classes import GSPath

        if self._path is not None:
            raise ValueError("Call endPath first.")

        self._path = GSPath()
        self._path.closed = True  # Until proven otherwise.

    def endPath(self) -> None:
        if self._path is None:
            raise ValueError("Call beginPath first.")

        if self._path.closed:
            self._path.nodes.append(self._path.nodes.pop(0))
        self._layer.paths.append(self._path)
        self._path = None

    def addPoint(
        self,
        pt: Tuple[Number, Number],
        segmentType: Optional[str] = None,
        smooth: bool = False,
        name: Optional[str] = None,
        userData: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> None:
        from glyphsLib.classes import GSNode

        if self._path is None:
            raise ValueError("Call beginPath first.")

        if segmentType == "move":
            if self._path.nodes:
                raise ValueError("For an open contour, 'move' must come first.")
            self._path.closed = False

        node = GSNode(
            Point(*pt),
            nodetype=_to_glyphs_node_type(segmentType),
            smooth=smooth,
            name=name,
        )
        if userData:
            node.userData = userData
        self._path.nodes.append(node)

    def addComponent(
        self,
        baseGlyph: str,
        transformation: Union[
            Transform, Tuple[float, float, float, float, float, float]
        ],
        **kwargs: Any,
    ) -> None:
        from glyphsLib.classes import GSComponent

        if not isinstance(transformation, Transform):
            transformation = Transform(*transformation)
        component = GSComponent(baseGlyph, transform=transformation)
        self._layer.components.append(component)


def _to_glyphs_node_type(node_type):
    if node_type is None:
        return "offcurve"
    if node_type == "move":
        return "line"
    return node_type


def _to_ufo_node_type(node_type):
    if node_type not in ["line", "curve", "qcurve"]:
        return None
    return node_type



"""Convert Glyphs smart components.

Smart components (https://glyphsapp.com/learn/smart-components) are a
feature within Glyphs whereby a component can essentially define its
own private designspace - each master of a component glyph can
define its own axes and masters. When the component is used, instead
of simply scaling or transforming the component, the designer can
*interpolate* the component by specifying the location in the private
designspace.

For example, a font might define a ``_part.serif`` glyph component with
"left width" and "right width" axes, and for each master in the font,
define layers for the ``_part.serif`` at some default left width and
right width, one with an extended left width, and one with an extended
right width, and use the "smart components settings" to assign locations
to these additional layers. (Unlike a full interpolation model, the
locations of smart component layers can only be at the axis extremes.)

We handle smart components by decomposing them and then applying a standard
OpenType interpolation model to adjust the node positions.
"""

from enum import IntEnum

from fontTools.varLib.models import VariationModel, normalizeValue, VariationModelError
from fontTools.ttLib.tables._g_l_y_f import GlyphCoordinates

from glyphsLib.classes import GSLayer


# smartComponentPoleMapping returns 1 for bottom of axis and 2 for top.
class Pole(IntEnum):
    MIN = 1
    MAX = 2


# This normalizes the location of a "master" (additional smart component
# layer). Because these are defined to be at the "poles", this is always
# 0.0, 1.0 or -1.0. But working out which it should be is slightly tricky:
# the axes don't define their own defaults, but the location of the
# default layer of the glyph tells you whether the default value of the
# axis is at the top or the bottom.
def normalized_location(layer, base_layer):
    loc = {}
    for axis_name, current_value in layer.smartComponentPoleMapping.items():
        base_value = base_layer.smartComponentPoleMapping[axis_name]
        if current_value == base_value:
            loc[axis_name] = 0.0
        elif base_value == Pole.MIN and current_value == Pole.MAX:
            loc[axis_name] = 1.0
        elif base_value == Pole.MAX and current_value == Pole.MIN:
            loc[axis_name] = -1.0
        else:
            raise ValueError(
                f"Strange axis mapping for axis {axis_name} in smart layer {base_layer}"
            )
    return loc


def variation_model(glyph, smart_layers, layer):
    master_locations = [normalized_location(l, smart_layers[0]) for l in smart_layers]
    axis_order = [ax.name for ax in glyph.smartComponentAxes]
    try:
        model = VariationModel(master_locations, axisOrder=axis_order, extrapolate=True)
    except VariationModelError as e:
        locations = "Locations were:\n"
        for smart_layer, master_location in zip(smart_layers, master_locations):
            locations += f"  {smart_layer.name} = {master_location}\n"
        raise ValueError(
            "Could not generate smart component model for %s used in %s.\n%s"
            % (glyph.name, layer, locations)
        ) from e
    return model


# Two slightly horrible functions for turning a GSLayer into a
# GlyphCoordinates object and back again.
def get_coordinates(layer):
    gc = GlyphCoordinates([])
    for path in layer.paths:
        gc.extend(
            GlyphCoordinates([(pt.position.x, pt.position.y) for pt in path.nodes])
        )
    return gc


def set_coordinates(layer, coords):
    counter = 0
    for path in layer.paths:
        for node in path.nodes:
            node.position.x, node.position.y = coords[counter]
            counter += 1


def decompose_smart_components_in_layer(self, layer):
    """Decompose any smart components in a layer, returning a new layer with paths.

    This recursively instantiates nested smart components before extracting
    coordinates for interpolation.

    See https://github.com/googlefonts/glyphsLib/issues/1111
    """
    from glyphsLib.pens import LayerPointPen

    has_smart_components = any(
        comp.component and comp.component.smartComponentAxes
        for comp in layer.components
    )
    if not has_smart_components:
        return layer

    from glyphsLib.classes import GSPath, GSComponent

    new_layer = GSLayer()
    new_layer.parent = layer.parent
    new_layer.associatedMasterId = layer.associatedMasterId

    pen = LayerPointPen(new_layer)
    for shape in layer._shapes:
        if isinstance(shape, GSPath):
            new_layer.paths.append(shape.clone())
        elif isinstance(shape, GSComponent):
            if shape.component and shape.component.smartComponentAxes:
                # Recursively decompose this smart component
                instantiate_smart_component(self, new_layer, shape, pen)
            else:
                pen.addComponent(shape.name, shape.transform)

    return new_layer


def get_smart_component_variation_model(layer, component):
    """Get the variation model for a smart component."""
    # Find the GSGlyph that is being used as a component by this GSComponent
    root = component.component

    masters = [l for l in root.layers if l.smartComponentPoleMapping]
    if layer.associatedMasterId:
        # Each master in the font can have its own set of smart component
        # "master layers", so we need to filter by those smart components
        # which are in the same font master as the current one
        masters = [
            l for l in masters if l.associatedMasterId == layer.associatedMasterId
        ]
    if not masters:
        raise ValueError(
            "Could not find any masters for the smart component %s used in %s"
            % (root.name, layer.name)
        )

    if len(masters) == 1:
        return None, None, None

    model = variation_model(root, masters, layer)

    # Determine the normalized location of the interpolant within the
    # mini-designspace, remembering that we have to work out where the
    # default value is by looking at the first "master"
    axes_tuples = {}
    for ax in root.smartComponentAxes:
        if masters[0].smartComponentPoleMapping[ax.name] == Pole.MIN:
            defaultValue = ax.bottomValue
        else:
            defaultValue = ax.topValue
        axes_tuples[ax.name] = (ax.bottomValue, defaultValue, ax.topValue)
    normalized_location = {
        name: normalizeValue(value, axes_tuples[name], extrapolate=True)
        for name, value in component.smartComponentValues.items()
    }

    return model, normalized_location, masters


def instantiate_smart_component(self, layer, component, pen):
    """Instantiate a smart component by interpolating and drawing to a pointPen."""
    # Find the GSGlyph that is being used as a component by this GSComponent
    root = component.component

    model, normalized_location, masters = get_smart_component_variation_model(
        layer, component
    )
    if model is None:
        # Treat this as a dumb component.
        pen.addComponent(component.name, component.transform)
        return

    # Decompose nested smart components before extracting coordinates
    decomposed_masters = [decompose_smart_components_in_layer(self, l) for l in masters]
    coordinates = [get_coordinates(l) for l in decomposed_masters]
    try:
        new_coords = model.interpolateFromMasters(normalized_location, coordinates)
    except Exception as e:
        raise ValueError(
            "Could not interpolate smart component %s used in %s" % (root.name, layer)
        ) from e

    # Decompose by creating a new layer, copying its shapes and applying
    # the new coordinates
    new_layer = GSLayer()
    new_layer._shapes = [shape.clone() for shape in decomposed_masters[0]._shapes]
    set_coordinates(new_layer, new_coords)

    # Don't forget that the GSComponent might also be transformed, so
    # we need to apply that transformation to the new layer as well
    if component.transform:
        # We must reverse path direction for flipped components
        # https://github.com/googlefonts/glyphsLib/issues/882
        should_reverse = component.transform.determinant() < 0
        for p in new_layer.paths:
            p.applyTransform(component.transform)
            if should_reverse:
                p.reverse()

    # And we are done
    new_layer.drawPoints(pen)



# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from glyphsLib.types import Point
import uuid

from glyphsLib.builder.constants import OBJECT_LIBS_KEY

__all__ = [
    "to_ufo_glyph_anchors",
    "to_glyphs_glyph_anchors",
]


def to_ufo_glyph_anchors(self, glyph, anchors):
    """Add .glyphs anchors to a glyph."""

    for anchor in anchors:
        x, y = anchor.position
        anchor_dict = {"name": anchor.name, "x": x, "y": y}
        if anchor.userData:
            identifier = str(uuid.uuid4()).upper()
            anchor_dict["identifier"] = identifier
            glyph.lib.setdefault(OBJECT_LIBS_KEY, {})[identifier] = dict(
                anchor.userData
            )
        glyph.appendAnchor(anchor_dict)


def to_glyphs_glyph_anchors(self, ufo_glyph, layer):
    """Add UFO glif anchors to a GSLayer."""
    for ufo_anchor in ufo_glyph.anchors:
        anchor = self.glyphs_module.GSAnchor()
        anchor.name = ufo_anchor.name
        anchor.position = Point(ufo_anchor.x, ufo_anchor.y)
        layer.anchors.append(anchor)



# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from .constants import CROP_KEY, LOCKED_KEY, ALPHA_KEY
from glyphsLib.types import Transform, Rect, Point, Size


def to_ufo_background_image(self, ufo_glyph, layer):
    """Copy the backgound image from the GSLayer to the UFO Glyph."""
    image = layer.backgroundImage
    if image is None:
        return
    ufo_image = ufo_glyph.image
    ufo_image.fileName = image.path
    ufo_image.transformation = image.transform
    ufo_glyph.lib[CROP_KEY] = list(image.crop)
    ufo_glyph.lib[LOCKED_KEY] = image.locked
    ufo_glyph.lib[ALPHA_KEY] = image.alpha


def to_glyphs_background_image(self, ufo_glyph, layer):
    """Copy the background image from the UFO Glyph to the GSLayer."""
    ufo_image = ufo_glyph.image
    if ufo_image.fileName is None:
        return
    image = self.glyphs_module.GSBackgroundImage()
    image.path = ufo_image.fileName
    image.transform = Transform(*ufo_image.transformation)
    if CROP_KEY in ufo_glyph.lib:
        x, y, w, h = ufo_glyph.lib[CROP_KEY]
        image.crop = Rect(Point(x, y), Size(w, h))
    if LOCKED_KEY in ufo_glyph.lib:
        image.locked = ufo_glyph.lib[LOCKED_KEY]
    if ALPHA_KEY in ufo_glyph.lib:
        image.alpha = ufo_glyph.lib[ALPHA_KEY]
    layer.backgroundImage = image



# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from .common import to_ufo_time, from_ufo_time
from .constants import (
    DEFAULT_FEATURE_WRITERS,
    UFO2FT_FEATURE_WRITERS_KEY,
    UFO2FT_FILTERS_KEY,
    APP_VERSION_LIB_KEY,
    KEYBOARD_INCREMENT_KEY,
    LANGUAGE_MAPPING,
    MASTER_ORDER_LIB_KEY,
)


def to_ufo_font_attributes(self, family_name):
    """Generate a list of UFOs with metadata loaded from .glyphs data.

    Modifies the list of UFOs in the UFOBuilder (self) in-place.
    """

    font = self.font

    for index, master in enumerate(font.masters):
        ufo = self.ufo_module.Font()

        fill_ufo_metadata(master, ufo)
        if not self.minimal:
            fill_ufo_metadata_roundtrip(master, ufo)

        self.to_ufo_names(ufo, master, family_name)  # .names
        self.to_ufo_family_user_data(ufo)  # .user_data

        if has_any_corner_components(font, master):
            ufo.lib.setdefault(UFO2FT_FILTERS_KEY, []).append(
                {
                    "namespace": "glyphsLib.filters",
                    "name": "cornerComponents",
                    "pre": True,
                }
            )

        ufo.lib.setdefault(UFO2FT_FILTERS_KEY, []).append(
            {"namespace": "glyphsLib.filters", "name": "eraseOpenCorners", "pre": True}
        )
        ufo.lib[UFO2FT_FEATURE_WRITERS_KEY] = DEFAULT_FEATURE_WRITERS

        self.to_ufo_custom_params(ufo, font)  # .custom_params
        self.to_ufo_master_attributes(ufo, master)  # .masters

        ufo.lib[MASTER_ORDER_LIB_KEY] = index

        # FIXME: (jany) in the future, yield this UFO (for memory, lazy iter)
        source = self._designspace.newSourceDescriptor()
        source.font = ufo
        self._designspace.addSource(source)
        self._sources[master.id] = source


INFO_FIELDS = (
    ("unitsPerEm", "upm"),
    ("versionMajor", "versionMajor"),
    ("versionMinor", "versionMinor"),
)

PROPERTIES_FIELDS = (
    ("copyright", "copyrights", 0),
    ("familyName", "familyNames", 1),
    ("trademark", "trademarks", 7),
    ("openTypeNameManufacturer", "manufacturers", 8),
    ("openTypeNameManufacturerURL", "manufacturerURL", None),
    ("openTypeNameDesigner", "designers", 9),
    ("openTypeNameDesignerURL", "designerURL", None),
    ("openTypeNameDescription", "descriptions", 10),
    ("openTypeNameLicense", "licenses", 13),
    ("openTypeNameSampleText", "sampleTexts", 19),
)


def fill_ufo_metadata(master, ufo):
    font = master.font

    for info_key, glyphs_key in INFO_FIELDS:
        value = getattr(font, glyphs_key)
        setattr(ufo.info, info_key, value)

    # font.properties will "helpfully" always give us teh property value as a
    # string, but we want the GSFontInfoValue to access its localized values
    properties = {p.key: p for p in font.properties}
    name_records = ufo.info.openTypeNameRecords or []
    for info_key, glyphs_key, name_id in PROPERTIES_FIELDS:
        value = properties.get(glyphs_key)
        if value:
            setattr(ufo.info, info_key, value.value)
            # If the property has localized values, add them to openTypeNameRecords
            if name_id is not None and value._localized_values:
                for lang, val in value._localized_values.items():
                    if lang in ["dflt", "default", "ENG"]:
                        continue
                    if (lang_id := LANGUAGE_MAPPING.get(lang)) is None:
                        raise ValueError(f"Unknown name language: {lang}")
                    name_records.append(
                        {
                            "nameID": name_id,
                            "languageID": lang_id,
                            "string": val,
                            "platformID": 3,  # Windows
                            "encodingID": 1,  # Unicode BMP
                        }
                    )
    if name_records:
        ufo.info.openTypeNameRecords = name_records

    # "date" can be missing; Glyphs.app removes it on saving if it's empty:
    # https://github.com/googlefonts/glyphsLib/issues/134
    date_created = getattr(font, "date", None)
    if date_created is not None:
        date_created = to_ufo_time(date_created)

    if date_created is not None:
        ufo.info.openTypeHeadCreated = date_created

    # NOTE: glyphs2ufo will *always* set a UFO public.glyphOrder equal to the
    # order of glyphs in the glyphs file, which can optionally be overwritten
    # by a glyphOrder custom parameter below in `to_ufo_custom_params`.
    ufo.glyphOrder = list(glyph.name for glyph in font.glyphs)


def fill_ufo_metadata_roundtrip(master, ufo):
    font = master.font
    ufo.lib[APP_VERSION_LIB_KEY] = font.appVersion
    ufo.lib[KEYBOARD_INCREMENT_KEY] = font.keyboardIncrement


# UFO to glyphs


def to_glyphs_font_attributes(self, source, master, is_initial):
    """
    Copy font attributes from `ufo` either to `self.font` or to `master`.

    Arguments:
    self -- The UFOBuilder
    ufo -- The current UFO being read
    master -- The current master being written
    is_initial -- True iff this the first UFO that we process
    """
    if is_initial:
        _set_glyphs_font_attributes(self, source)
    else:
        _compare_and_merge_glyphs_font_attributes(self, source)


def _set_glyphs_font_attributes(self, source):
    font = self.font
    ufo = source.font
    info = ufo.info

    if APP_VERSION_LIB_KEY in ufo.lib:
        font.appVersion = ufo.lib[APP_VERSION_LIB_KEY]
    if KEYBOARD_INCREMENT_KEY in ufo.lib:
        font.keyboardIncrement = ufo.lib[KEYBOARD_INCREMENT_KEY]

    if info.openTypeHeadCreated is not None:
        # FIXME: (jany) should wrap in glyphs_datetime? or maybe the GSFont
        #     should wrap in glyphs_datetime if needed?
        font.date = from_ufo_time(info.openTypeHeadCreated)
    if info.unitsPerEm is not None:
        font.upm = info.unitsPerEm
    if info.versionMajor is not None:
        font.versionMajor = info.versionMajor
    if info.versionMinor is not None:
        font.versionMinor = info.versionMinor

    if info.copyright is not None:
        font.copyright = info.copyright
    if info.openTypeNameDesigner is not None:
        font.designer = info.openTypeNameDesigner
    if info.openTypeNameDesignerURL is not None:
        font.designerURL = info.openTypeNameDesignerURL
    if info.openTypeNameManufacturer is not None:
        font.manufacturer = info.openTypeNameManufacturer
    if info.openTypeNameManufacturerURL is not None:
        font.manufacturerURL = info.openTypeNameManufacturerURL

    self.to_glyphs_family_names(ufo)
    self.to_glyphs_family_user_data_from_ufo(ufo)
    self.to_glyphs_custom_params(ufo, font)


def _compare_and_merge_glyphs_font_attributes(self, source):
    ufo = source.font
    self.to_glyphs_family_names(ufo, merge=True)


def to_glyphs_ordered_masters(self):
    """Modify in-place the list of UFOs to restore their original order in
    the Glyphs file (if any, otherwise does not change the order)."""
    return sorted(self.designspace.sources, key=_original_master_order)


def _original_master_order(source):
    try:
        return source.font.lib[MASTER_ORDER_LIB_KEY]
    # Key may not be found or source.font be None if it's a layer source.
    except (KeyError, AttributeError):
        return 1 << 31


def has_any_corner_components(font, master):
    for glyph in font.glyphs:
        for layer in glyph.layers:
            if (
                layer.layerId != master.id
                or layer.associatedMasterId != master.id
                or not layer.hints
            ):
                continue
            if any(h.type.upper() == "CORNER" for h in layer.hints):
                return True
    return False



# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import logging
import os
import posixpath

from .constants import (
    GLYPHS_PREFIX,
    PUBLIC_PREFIX,
    UFO2FT_FEATURE_WRITERS_KEY,
    DEFAULT_FEATURE_WRITERS,
    DEFAULT_LAYER_NAME,
    UFO_DATA_KEY,
    FONT_USER_DATA_KEY,
    LAYER_LIB_KEY,
    LAYER_NAME_KEY,
    GLYPH_USER_DATA_KEY,
    NODE_USER_DATA_KEY,
    GLYPHS_MATH_VARIANTS_KEY,
    GLYPHS_MATH_EXTENDED_SHAPE_KEY,
    GLYPHS_MATH_PREFIX,
)

logger = logging.getLogger(__name__)


def to_designspace_family_user_data(self):
    if self.use_designspace:
        for key, value in dict(self.font.userData).items():
            if _user_data_has_no_special_meaning(key):
                self.designspace.lib[key] = value


def to_ufo_family_user_data(self, ufo):
    """Set family-wide user data as Glyphs does."""
    if not self.use_designspace:
        ufo.lib[FONT_USER_DATA_KEY] = dict(self.font.userData)


def to_ufo_master_user_data(self, ufo, master):
    """Set master-specific user data as Glyphs does."""
    for key in master.userData.keys():
        if _user_data_has_no_special_meaning(key):
            ufo.lib[key] = master.userData[key]

    # Restore UFO data files. This code assumes that all paths are POSIX paths.
    if UFO_DATA_KEY in master.userData:
        for filename, data in master.userData[UFO_DATA_KEY].items():
            ufo.data[filename] = bytes(data)


def to_ufo_glyph_user_data(self, ufo, ufo_glyph, glyph):
    math_data = {
        k: v for k, v in glyph.userData.items() if k.startswith(GLYPHS_MATH_PREFIX)
    }
    if math_data:
        # Convert MATH userData to top-level keys and group them under the same
        # key so that they are in a more usable/compact form.
        if GLYPHS_MATH_EXTENDED_SHAPE_KEY in math_data:
            ufo.lib.setdefault(GLYPHS_MATH_EXTENDED_SHAPE_KEY, []).append(glyph.name)
        if GLYPHS_MATH_VARIANTS_KEY in math_data:
            ufo_glyph.lib.setdefault(GLYPHS_MATH_VARIANTS_KEY, {}).update(
                dict(math_data[GLYPHS_MATH_VARIANTS_KEY])
            )
    if self.minimal:
        return
    other_data = {k: v for k, v in glyph.userData.items() if k not in math_data}
    key = GLYPH_USER_DATA_KEY + "." + glyph.name
    if other_data:
        ufo.lib[key] = dict(other_data)


def to_ufo_layer_lib(self, master, ufo, ufo_layer):
    key = LAYER_LIB_KEY + "." + ufo_layer.name
    # glyphsLib v5.3.2 and previous versions stored the layer lib in
    # the GSFont useData under a key named after the layer.
    # When different original UFOs each had a layer with the same layer name,
    # only the layer lib of the last one was stored and was exported to UFOs
    if key in self.font.userData.keys():
        ufo_layer.lib.update(self.font.userData[key])
    if key in master.userData.keys():
        ufo_layer.lib.update(master.userData[key])
        if LAYER_NAME_KEY in ufo_layer.lib:
            layer_name = ufo_layer.lib.pop(LAYER_NAME_KEY)
            # ufoLib2
            if hasattr(ufo, "renameLayer") and callable(ufo.renameLayer):
                ufo.renameLayer(ufo_layer.name, layer_name)
            # defcon
            else:
                ufo_layer.name = layer_name


def to_ufo_layer_user_data(self, ufo_glyph, layer):
    user_data = layer.userData
    math_data = {k: v for k, v in user_data.items() if k.startswith(GLYPHS_MATH_PREFIX)}
    if math_data:
        if GLYPHS_MATH_VARIANTS_KEY in math_data:
            ufo_glyph.lib.setdefault(GLYPHS_MATH_VARIANTS_KEY, {}).update(
                dict(math_data[GLYPHS_MATH_VARIANTS_KEY])
            )
    if self.minimal:
        return
    other_data = {k: v for k, v in user_data.items() if k not in math_data}
    for key in other_data.keys():
        if _user_data_has_no_special_meaning(key):
            ufo_glyph.lib[key] = other_data[key]


def to_ufo_node_user_data(self, ufo_glyph, node, user_data: dict):
    if user_data:
        path_index, node_index = node._indices()
        key = f"{NODE_USER_DATA_KEY}.{path_index}.{node_index}"
        ufo_glyph.lib[key] = user_data


def to_glyphs_family_user_data_from_designspace(self):
    """Set the GSFont userData from the designspace family-wide lib data."""
    target_user_data = self.font.userData
    for key, value in self.designspace.lib.items():
        if key == UFO2FT_FEATURE_WRITERS_KEY and value == DEFAULT_FEATURE_WRITERS:
            # if the designspace contains featureWriters settings that are the
            # same as glyphsLib default settings, there's no need to store them
            continue
        if _user_data_has_no_special_meaning(key):
            target_user_data[key] = value


def to_glyphs_family_user_data_from_ufo(self, ufo):
    """Set the GSFont userData from the UFO family-wide lib data."""
    target_user_data = self.font.userData
    try:
        for key, value in ufo.lib[FONT_USER_DATA_KEY].items():
            # Existing values taken from the designspace lib take precedence
            if key not in target_user_data.keys():
                target_user_data[key] = value
    except KeyError:
        # No FONT_USER_DATA in ufo.lib
        pass


def to_glyphs_master_user_data(self, ufo, master):
    """Set the GSFontMaster userData from the UFO master-specific lib data."""
    target_user_data = master.userData
    special_math_keys = {GLYPHS_MATH_VARIANTS_KEY, GLYPHS_MATH_EXTENDED_SHAPE_KEY}
    for key, value in ufo.lib.items():
        if _user_data_has_no_special_meaning(key) and key not in special_math_keys:
            target_user_data[key] = value

    # Save UFO data files
    if ufo.data.fileNames:
        from glyphsLib.types import BinaryData

        ufo_data = {}
        for os_filename in ufo.data.fileNames:
            filename = posixpath.join(*os_filename.split(os.path.sep))
            ufo_data[filename] = BinaryData(ufo.data[os_filename])
        master.userData[UFO_DATA_KEY] = ufo_data


def to_glyphs_glyph_user_data(self, ufo, glyph):
    key = GLYPH_USER_DATA_KEY + "." + glyph.name
    if key in ufo.lib:
        glyph.userData = ufo.lib[key]

    key = GLYPHS_MATH_EXTENDED_SHAPE_KEY
    if key in ufo.lib and glyph.name in ufo.lib[key]:
        glyph.userData[key] = True

    key = GLYPHS_MATH_VARIANTS_KEY
    if key in ufo.lib and glyph.name in ufo.lib[key]:
        glyph.userData[key] = ufo.lib[key][glyph.name]


def to_glyphs_layer_lib(self, ufo_layer, master):
    user_data = {}
    for key, value in ufo_layer.lib.items():
        if _user_data_has_no_special_meaning(key):
            user_data[key] = value

    # the default layer may have a custom name
    layer_name = ufo_layer.name
    if (
        ufo_layer is self._sources[master.id].font.layers.defaultLayer
        and layer_name != DEFAULT_LAYER_NAME
    ):
        user_data[LAYER_NAME_KEY] = ufo_layer.name
        layer_name = DEFAULT_LAYER_NAME

    if user_data:
        key = LAYER_LIB_KEY + "." + layer_name
        master.userData[key] = user_data


def to_glyphs_layer_user_data(self, ufo_glyph, layer):
    user_data = layer.userData
    for key, value in ufo_glyph.lib.items():
        if key == GLYPHS_MATH_VARIANTS_KEY:
            glyph_user_data = layer.parent.userData
            for sub_key, sub_value in value.items():
                if sub_key in {"vVariants", "hVariants"}:
                    if key not in glyph_user_data:
                        glyph_user_data[key] = {}
                    print(sub_value, glyph_user_data[key].get(sub_key))
                    if (
                        sub_key in glyph_user_data[key]
                        and glyph_user_data[key][sub_key] != sub_value
                    ):
                        logger.warning(
                            f"Glyph '{layer.parent.name}' already has different "
                            f"'{sub_key}' in userData['{key}']. Overwriting it."
                        )
                    glyph_user_data[key][sub_key] = sub_value
                else:
                    if key not in user_data:
                        user_data[key] = {}
                    user_data[key][sub_key] = sub_value
        elif _user_data_has_no_special_meaning(key):
            user_data[key] = value


def to_glyphs_node_user_data(self, ufo_glyph, node, path_index, node_index):
    key = f"{NODE_USER_DATA_KEY}.{path_index}.{node_index}"
    if key in ufo_glyph.lib:
        for k, v in ufo_glyph.lib[key].items():
            if k == "name":
                continue  # We take the node name from a UFO point's name attribute.
            node.userData[k] = v


def _user_data_has_no_special_meaning(key):
    return not (key.startswith(GLYPHS_PREFIX) or key.startswith(PUBLIC_PREFIX))



#
# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest
import datetime
import os

from glyphsLib.builder.constants import GLYPHS_COLORS, GLYPHLIB_PREFIX
from glyphsLib import to_glyphs, to_ufos, to_designspace
from glyphsLib import classes

from fontTools.designspaceLib import DesignSpaceDocument, AxisDescriptor

# TODO: (jany) think hard about the ordering and RTL/LTR
# TODO: (jany) make one generic test with data using pytest


@pytest.mark.skip
def test_anchors_with_same_name_correct_order_rtl(ufo_module):
    ufo = ufo_module.Font()
    g = ufo.newGlyph("laam_alif")
    # Append the anchors in the correct order
    g.appendAnchor(dict(x=50, y=600, name="top"))
    g.appendAnchor(dict(x=250, y=600, name="top"))

    font = to_glyphs([ufo])

    top1, top2 = font.glyphs["laam_alif"].layers[0].anchors

    assert top1.name == "top_1"
    assert top1.x == 50
    assert top1.y == 600
    assert top2.name == "top_2"
    assert top2.x == 250
    assert top2.y == 600


@pytest.mark.skip
def test_anchors_with_same_name_wrong_order_rtl(ufo_module):
    ufo = ufo_module.Font()
    g = ufo.newGlyph("laam_alif")
    # Append the anchors in the wrong order
    g.appendAnchor(dict(x=250, y=600, name="top"))
    g.appendAnchor(dict(x=50, y=600, name="top"))

    font = to_glyphs([ufo])

    top1, top2 = font.glyphs["laam_alif"].layers[0].anchors

    # FIXME: (jany) think hard about the ordering and LTR
    assert top1.name == "top_1"
    assert top1.x == 50
    assert top1.y == 600
    assert top2.name == "top_2"
    assert top2.x == 250
    assert top2.y == 600


@pytest.mark.skip
def test_anchors_with_same_name_correct_order_ltr(ufo_module):
    ufo = ufo_module.Font()
    g = ufo.newGlyph("laam_alif")
    # Append the anchors in the correct order
    g.appendAnchor(dict(x=50, y=600, name="top"))
    g.appendAnchor(dict(x=250, y=600, name="top"))

    font = to_glyphs([ufo])

    top1, top2 = font.glyphs["laam_alif"].layers[0].anchors

    # FIXME: (jany) think hard about the ordering and RTL/LTR
    assert top1.name == "top_1"
    assert top1.x == 50
    assert top1.y == 600
    assert top2.name == "top_2"
    assert top2.x == 250
    assert top2.y == 600


@pytest.mark.skip
def test_anchors_with_same_name_wrong_order_ltr(ufo_module):
    ufo = ufo_module.Font()
    g = ufo.newGlyph("laam_alif")
    # Append the anchors in the wrong order
    g.appendAnchor(dict(x=250, y=600, name="top"))
    g.appendAnchor(dict(x=50, y=600, name="top"))

    font = to_glyphs([ufo])

    top1, top2 = font.glyphs["laam_alif"].layers[0].anchors

    # FIXME: (jany) think hard about the ordering and LTR
    assert top1.name == "top_1"
    assert top1.x == 50
    assert top1.y == 600
    assert top2.name == "top_2"
    assert top2.x == 250
    assert top2.y == 600


def test_groups(ufo_module):
    ufo = ufo_module.Font()
    ufo.newGlyph("T")
    ufo.newGlyph("e")
    ufo.newGlyph("o")
    samekh = ufo.newGlyph("samekh-hb")
    samekh.unicode = 0x05E1
    resh = ufo.newGlyph("resh-hb")
    resh.unicode = 0x05E8
    ufo.groups["public.kern1.T"] = ["T"]
    ufo.groups["public.kern2.oe"] = ["o", "e"]
    ufo.groups["com.whatever.Te"] = ["T", "e"]
    # Groups can contain glyphs that are not in the font and that should
    # be preserved as well
    ufo.groups["public.kern1.notInFont"] = ["L"]
    ufo.groups["public.kern1.halfInFont"] = ["o", "b", "p"]
    ufo.groups["com.whatever.notInFont"] = ["i", "j"]
    # Empty groups as well (found in the wild)
    ufo.groups["public.kern1.empty"] = []
    ufo.groups["com.whatever.empty"] = []
    # Groups for RTL glyphs. In a UFO RTL kerning pair, kern1 is for the glyph
    # on the left visually (the first that gets written when writing RTL)
    # The example below with Resh and Samekh comes from:
    # https://forum.glyphsapp.com/t/dramatic-bug-in-hebrew-kerning/4093
    ufo.groups["public.kern1.hebrewLikeT"] = ["resh-hb"]
    ufo.groups["public.kern2.hebrewLikeO"] = ["samekh-hb"]
    groups_dict = dict(ufo.groups)

    # TODO: (jany) add a test with 2 UFOs with conflicting data
    # TODO: (jany) add a test with with both UFO groups and feature file classes
    # TODO: (jany) add a test with UFO groups that conflict with feature file classes
    font = to_glyphs([ufo], minimize_ufo_diffs=True)

    # Kerning for existing glyphs is stored in GSGlyph.left/rightKerningGroup
    assert font.glyphs["T"].rightKerningGroup == "T"
    assert font.glyphs["o"].leftKerningGroup == "oe"
    assert font.glyphs["e"].leftKerningGroup == "oe"
    # ...and actually in Glyphs, the "rightKerningGroup" is NOT always on the
    # right, but always matches "public.kern1" in UFO, so there is nothing
    # special to do for RTL.
    assert font.glyphs["resh-hb"].rightKerningGroup == "hebrewLikeT"
    assert font.glyphs["samekh-hb"].leftKerningGroup == "hebrewLikeO"

    # Non-kerning groups are stored as classes
    assert font.classes["com.whatever.Te"].code == "T e"
    assert font.classes["com.whatever.notInFont"].code == "i j"
    # Kerning groups with some characters not in the font are also saved
    # somehow, but we don't care how, that fact will be better tested by the
    # roundtrip test a few lines below

    (ufo,) = to_ufos(font)

    # Check that nothing has changed
    assert dict(ufo.groups) == groups_dict

    # Check that changing the `left/rightKerningGroup` fields in Glyphs
    # updates the UFO kerning groups
    font.glyphs["T"].rightKerningGroup = "newNameT"
    font.glyphs["o"].rightKerningGroup = "onItsOwnO"

    del groups_dict["public.kern1.T"]
    groups_dict["public.kern1.newNameT"] = ["T"]
    groups_dict["public.kern1.halfInFont"].remove("o")
    groups_dict["public.kern1.onItsOwnO"] = ["o"]

    (ufo,) = to_ufos(font)

    assert dict(ufo.groups) == groups_dict


def test_guidelines(ufo_module):
    ufo = ufo_module.Font()
    a = ufo.newGlyph("a")
    for obj in [ufo, a]:
        # Complete guideline
        obj.appendGuideline(
            dict(x=10, y=20, angle=30, name="lc", color="1,0,0,1", identifier="lc1")
        )
        # Don't crash if a guideline misses information
        obj.appendGuideline({"x": 10})
        obj.appendGuideline({"y": 20})

    font = to_glyphs([ufo])

    for gobj in [font.masters[0], font.glyphs["a"].layers[0]]:
        assert len(gobj.guides) == 3

        angled, vertical, horizontal = gobj.guides

        assert angled.position.x == 10
        assert angled.position.y == 20
        assert angled.angle == 30
        assert angled.name == "lc [1,0,0,1] [#lc1]"

        assert vertical.position.x == 10
        assert vertical.angle == 90

        assert horizontal.position.y == 20
        assert horizontal.angle == 0

    (ufo,) = to_ufos(font, minimal=False)

    for obj in [ufo, ufo["a"]]:
        angled, vertical, horizontal = obj.guidelines

        assert angled.x == 10
        assert angled.y == 20
        assert angled.angle == 30
        assert angled.name == "lc"
        assert angled.color == "1,0,0,1"
        assert angled.identifier == "lc1"

        assert vertical.x == 10
        assert vertical.y is None
        assert vertical.angle is None

        assert horizontal.x is None
        assert horizontal.y == 20
        assert horizontal.angle is None


def test_glyph_color(ufo_module):
    ufo = ufo_module.Font()
    a = ufo.newGlyph("a")
    a.markColor = GLYPHS_COLORS[3]
    b = ufo.newGlyph("b")
    b.markColor = "{:.3f},{:.3f},0,1".format(4.0 / 255, 128.0 / 255)

    font = to_glyphs([ufo])

    rt_a_color = font.glyphs["a"].color
    assert rt_a_color == 3 and type(rt_a_color) is int
    rt_b_color = font.glyphs["b"].color
    assert rt_b_color == [4, 128, 0, 255] and all(type(x) is int for x in rt_b_color)

    (ufo,) = to_ufos(font)

    assert ufo["a"].markColor == GLYPHS_COLORS[3]
    assert ufo["b"].markColor == b.markColor


def test_glyph_height_and_vertical_origin(ufo_module):
    ufo = ufo_module.Font()
    ufo.info.ascender = 1000
    ufo.info.descender = -200
    ufo.info.unitsPerEm = 1000

    a = ufo.newGlyph("a")
    a.height = 1024
    a.verticalOrigin = 100
    b = ufo.newGlyph("b")
    b.height = ufo.info.ascender - ufo.info.descender  # Glyphs default vertWidth
    c = ufo.newGlyph("c")
    assert c.height == 0  # defcon default glyph.height == 0

    font = to_glyphs([ufo])

    assert font.glyphs["a"].layers[0].vertWidth == ufo["a"].height
    assert font.glyphs["a"].layers[0].vertOrigin == 900
    assert font.glyphs["b"].layers[0].vertWidth is None
    assert font.glyphs["b"].layers[0].vertOrigin is None
    assert font.glyphs["c"].layers[0].vertWidth == 0
    assert font.glyphs["c"].layers[0].vertOrigin is None

    (ufo,) = to_ufos(font)

    assert ufo["a"].height == font.glyphs["a"].layers[0].vertWidth
    assert ufo["a"].verticalOrigin == 100
    assert ufo["b"].height == (ufo.info.ascender - ufo.info.descender)
    assert ufo["b"].verticalOrigin == ufo.info.ascender
    assert ufo["c"].height == 0
    assert ufo["c"].verticalOrigin == ufo.info.ascender


def test_glyph_height_and_vertical_origin_trigger_vhea_metrics(ufo_module):
    ufo = ufo_module.Font()
    ufo.info.ascender = 1000
    ufo.info.descender = -200
    ufo.info.unitsPerEm = 1000

    a = ufo.newGlyph("a")
    a.height = 1024
    a.verticalOrigin = 100

    font = to_glyphs([ufo])

    assert all(
        "vhea" not in custom_param.name
        for custom_param in font.masters[0].customParameters
    )

    (ufo,) = to_ufos(font)

    assert ufo.info.openTypeVheaVertTypoAscender == 500
    assert ufo.info.openTypeVheaVertTypoDescender == -500
    assert ufo.info.openTypeVheaVertTypoLineGap == 1000


def test_bad_ufo_date_format_in_glyph_lib(ufo_module):
    ufo = ufo_module.Font()
    a = ufo.newGlyph("a")
    a.lib[GLYPHLIB_PREFIX + "lastChange"] = "2017-12-19 15:12:44 +0000"

    # Don't crash
    font = to_glyphs([ufo])

    assert font.glyphs["a"].lastChange == datetime.datetime(2017, 12, 19, 15, 12, 44)


def test_have_default_interpolation_values(ufo_module):
    """When no designspace is provided, make sure that the Glyphs file has some
    default "axis positions" for the masters.
    """
    thin = ufo_module.Font()
    thin.info.openTypeOS2WidthClass = 5
    thin.info.openTypeOS2WeightClass = 100
    regular = ufo_module.Font()
    regular.info.openTypeOS2WidthClass = 5
    regular.info.openTypeOS2WeightClass = 400
    bold = ufo_module.Font()
    bold.info.openTypeOS2WidthClass = 5
    bold.info.openTypeOS2WeightClass = 700
    thin_expanded = ufo_module.Font()
    thin_expanded.info.openTypeOS2WidthClass = 7
    thin_expanded.info.openTypeOS2WeightClass = 100
    bold_ultra_cond = ufo_module.Font()
    bold_ultra_cond.info.openTypeOS2WidthClass = 1
    bold_ultra_cond.info.openTypeOS2WeightClass = 700

    font = to_glyphs([thin, regular, bold, thin_expanded, bold_ultra_cond])

    gthin, greg, gbold, gthinex, gbolducond = font.masters

    # For weight, copy the WeightClass as-is
    assert gthin.weightValue == 100
    assert greg.weightValue == 400
    assert gbold.weightValue == 700
    assert gthinex.weightValue == 100
    assert gbolducond.weightValue == 700

    # For width, use the "% of normal" column from the spec
    # https://www.microsoft.com/typography/otspec/os2.htm#wdc
    assert gthin.widthValue == 100
    assert greg.widthValue == 100
    assert gbold.widthValue == 100
    assert gthinex.widthValue == 125
    assert gbolducond.widthValue == 50


def test_designspace_source_locations(tmpdir, ufo_module):
    """Check that opening UFOs from their source descriptor works with both
    the filename and the path attributes.
    """
    designspace_path = os.path.join(str(tmpdir), "test.designspace")
    light_ufo_path = os.path.join(str(tmpdir), "light.ufo")
    bold_ufo_path = os.path.join(str(tmpdir), "bold.ufo")

    designspace = DesignSpaceDocument()
    wght = AxisDescriptor()
    wght.minimum = 100
    wght.maximum = 700
    wght.default = 100
    wght.name = "Weight"
    wght.tag = "wght"
    designspace.addAxis(wght)
    light_source = designspace.newSourceDescriptor()
    light_source.filename = "light.ufo"
    light_source.location = {"Weight": 100}
    designspace.addSource(light_source)
    bold_source = designspace.newSourceDescriptor()
    bold_source.path = bold_ufo_path
    bold_source.location = {"Weight": 700}
    designspace.addSource(bold_source)
    designspace.write(designspace_path)

    light = ufo_module.Font()
    light.info.ascender = 30
    light.save(light_ufo_path)

    bold = ufo_module.Font()
    bold.info.ascender = 40
    bold.save(bold_ufo_path)

    designspace = DesignSpaceDocument()
    designspace.read(designspace_path)

    font = to_glyphs(designspace, ufo_module=ufo_module)

    assert len(font.masters) == 2
    assert font.masters[0].ascender == 30
    assert font.masters[1].ascender == 40


@pytest.mark.skip(reason="Should be better defined")
def test_ufo_filename_is_kept_the_same(tmpdir, ufo_module):
    """Check that the filenames of existing UFOs are correctly written to
    the designspace document when doing UFOs -> Glyphs -> designspace.
    This only works when the option "minimize_ufo_diffs" is given, because
    keeping track of this information adds stuff to the Glyphs file.
    """
    light_ufo_path = os.path.join(str(tmpdir), "light.ufo")
    bold_ufo_path = os.path.join(str(tmpdir), "subdir/bold.ufo")

    light = ufo_module.Font()
    light.info.ascender = 30
    light.save(light_ufo_path)

    bold = ufo_module.Font()
    bold.info.ascender = 40
    bold.save(bold_ufo_path)

    # First check: when going from UFOs -> Glyphs -> designspace
    font = to_glyphs([light, bold], minimize_ufo_diffs=True)

    designspace = to_designspace(font, ufo_module=ufo_module)
    assert designspace.sources[0].path == light_ufo_path
    assert designspace.sources[1].path == bold_ufo_path

    # Second check: going from designspace -> Glyphs -> designspace
    designspace_path = os.path.join(str(tmpdir), "test.designspace")
    designspace = DesignSpaceDocument()
    light_source = designspace.newSourceDescriptor()
    light_source.filename = "light.ufo"
    designspace.addSource(light_source)
    bold_source = designspace.newSourceDescriptor()
    bold_source.path = bold_ufo_path
    designspace.addSource(bold_source)
    designspace.write(designspace_path)

    font = to_glyphs([light, bold], minimize_ufo_diffs=True)

    assert designspace.sources[0].filename == "light.ufo"
    assert designspace.sources[1].filename == "subdir/bold.ufo"


def test_dont_copy_advance_to_the_background_unless_it_was_there(tmpdir, ufo_module):
    ufo = ufo_module.Font()
    bg = ufo.newLayer("public.background")

    fg_a = ufo.newGlyph("a")
    fg_a.width = 100
    bg.newGlyph("a")

    fg_b = ufo.newGlyph("b")
    fg_b.width = 200
    bg_b = bg.newGlyph("b")
    bg_b.width = 300

    fg_c = ufo.newGlyph("c")
    fg_c.width = 400
    bg_c = bg.newGlyph("c")
    bg_c.width = 400

    converted_font = to_glyphs([ufo])
    path = os.path.join(str(tmpdir), "test.glyphs")
    converted_font.save(path)
    saved_font = classes.GSFont(path)

    for font in [converted_font, saved_font]:
        (ufo,) = to_ufos(font, minimal=False)

        assert ufo["a"].width == 100
        assert ufo.layers["public.background"]["a"].width == 0
        assert ufo["b"].width == 200
        assert ufo.layers["public.background"]["b"].width == 300
        assert ufo["c"].width == 400
        assert ufo.layers["public.background"]["c"].width == 400


def test_dont_zero_width_of_nonspacing_marks_if_it_was_not_zero():
    # TODO
    pass


def test_double_unicodes(tmpdir, ufo_module):
    ufo = ufo_module.Font()
    z = ufo.newGlyph("z")
    z.unicodes = [0x005A, 0x007A]

    converted_font = to_glyphs([ufo])
    path = os.path.join(str(tmpdir), "test.glyphs")
    converted_font.save(path)
    saved_font = classes.GSFont(path)

    for font in [converted_font, saved_font]:
        assert font.glyphs["z"].unicode == "005A"
        assert font.glyphs["z"].unicodes == ["005A", "007A"]

        (ufo,) = to_ufos(font)

        assert ufo["z"].unicodes == [0x005A, 0x007A]


def test_open_contour(ufo_module):
    ufo = ufo_module.Font()
    a = ufo.newGlyph("a")
    pen = a.getPen()
    pen.moveTo((10, 20))
    pen.lineTo((30, 40))
    pen.endPath()

    font = to_glyphs([ufo])

    path = font.glyphs["a"].layers[0].paths[0]
    assert not path.closed
    assert len(path.nodes) == 2
    assert path.nodes[0].type == classes.LINE

    (ufo_rt,) = to_ufos(font)

    assert [(p.segmentType, p.x, p.y) for p in a[0]] == [
        (p.segmentType, p.x, p.y) for p in ufo_rt["a"][0]
    ]


def test_background_before_foreground(ufo_module):
    ufo = ufo_module.Font()
    ufo.newGlyph("a")
    background = ufo.newLayer("public.background")
    background.newGlyph("a")

    ufo.layers.layerOrder = ["public.background", "public.default"]

    # Check that it does not crash
    to_glyphs([ufo])


def test_only_background(ufo_module):
    ufo = ufo_module.Font()
    background = ufo.newLayer("public.background")
    background.newGlyph("a")

    # Check that it does not crash
    to_glyphs([ufo])


def test_warn_diff_between_designspace_and_ufos(caplog, ufo_module):
    ufo = ufo_module.Font()
    ufo.info.familyName = "UFO Family Name"
    ufo.info.styleName = "UFO Style Name"
    # ufo.info.styleMapFamilyName = 'UFO Stylemap Family Name'
    # ufo.info.styleMapStyleName = 'bold'

    doc = DesignSpaceDocument()
    source = doc.newSourceDescriptor()
    source.font = ufo
    source.familyName = "DS Family Name"
    source.styleName = "DS Style Name"
    doc.addSource(source)

    font = to_glyphs(doc, minimize_ufo_diffs=True)
    assert any(record.levelname == "WARNING" for record in caplog.records)
    assert (
        "The familyName is different between the UFO and the designspace source"
        in caplog.text
    )
    assert (
        "The styleName is different between the UFO and the designspace source"
        in caplog.text
    )

    doc = to_designspace(font, ufo_module=ufo_module)
    source = doc.sources[0]

    # The UFO info will prevail
    assert ufo.info.familyName == "UFO Family Name"
    assert ufo.info.styleName == "UFO Style Name"
    assert source.font.info.familyName == "UFO Family Name"
    assert source.font.info.styleName == "UFO Style Name"


def test_custom_stylemap_style_name(ufo_module):
    ufo = ufo_module.Font()
    ufo.info.styleMapStyleName = "bold"  # Not "regular"

    font = to_glyphs([ufo], minimize_ufo_diffs=True)
    (ufo,) = to_ufos(font)

    assert ufo.info.styleMapStyleName == "bold"


def test_custom_default_layer_name(ufo_module):
    ufo1 = ufo_module.Font()
    ufo2 = ufo_module.Font()
    if hasattr(ufo1, "renameLayer") and callable(ufo1.renameLayer):
        ufo1.renameLayer("public.default", "custom default")
        ufo2.renameLayer("public.default", "other default")
    else:
        ufo1.layers.defaultLayer.name = "custom default"
        ufo2.layers.defaultLayer.name = "other default"

    font = to_glyphs([ufo1, ufo2])
    (ufo1, ufo2) = to_ufos(font)

    assert ufo1.layers.defaultLayer.name == "custom default"
    assert ufo2.layers.defaultLayer.name == "other default"


def test_weird_kerning_roundtrip(ufo_module):
    groups = {
        "public.kern1.i": [
            "i",
            "dotlessi",
            "iacute",
            "icircumflex",
            "idieresis",
            "igrave",
            "imacron",
            "iogonek",
            "itilde",
            "f_i",
            "f_f_i",
            "fi",
            "ffi",
        ],
        "public.kern2.i": [
            "i",
            "dotlessi",
            "iacute",
            "icircumflex",
            "idieresis",
            "igrave",
            "ij",
            "imacron",
            "iogonek",
            "itilde",
            "dotlessij",
            "ijacute",
        ],
    }
    kerning = {
        ("icircumflex", "public.kern2.i"): 20,
        ("idieresis", "idieresis"): 125,
        ("idieresis", "public.kern2.i"): 35,
        ("itilde", "public.kern2.i"): 10,
        ("public.kern1.i", "icircumflex"): 15,
        ("public.kern1.i", "idieresis"): 40,
        ("public.kern1.i", "itilde"): 10,
    }
    glyphs = sorted({g for glyphs in groups.values() for g in glyphs})

    ufo = ufo_module.Font()
    for glyph in glyphs:
        ufo.newGlyph(glyph)
    for k, v in groups.items():
        ufo.groups[k] = v
    for k, v in kerning.items():
        ufo.kerning[k] = v

    font = to_glyphs([ufo])
    (ufo,) = to_ufos(font)

    # The kerning and groups should round-trip untouched
    for name, glyphs in groups.items():
        assert set(glyphs) == set(ufo.groups[name])

    assert ufo.kerning == kerning



#
# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import glyphsLib
from fontTools.designspaceLib import DesignSpaceDocument
from glyphsLib.builder.instances import apply_instance_data

import pytest
import py.path
from ..test_helpers import write_designspace_and_UFOs


DATA = os.path.join(os.path.dirname(os.path.dirname(__file__)), "data")


@pytest.mark.parametrize(
    "instance_names",
    [None, ["Extra Light"], ["Regular", "Bold"]],
    ids=["default", "include_1", "include_2"],
)
def test_apply_instance_data(tmpdir, instance_names, ufo_module):
    font = glyphsLib.GSFont(os.path.join(DATA, "GlyphsUnitTestSans.glyphs"))
    instance_dir = "instances"
    designspace = glyphsLib.to_designspace(font, instance_dir=instance_dir)
    path = str(tmpdir / (font.familyName + ".designspace"))
    write_designspace_and_UFOs(designspace, path)

    test_designspace = DesignSpaceDocument()
    test_designspace.read(designspace.path)
    if instance_names is None:
        # Collect all instances.
        test_instances = [instance.filename for instance in test_designspace.instances]
    else:
        # Collect only selected instances.
        test_instances = [
            instance.filename
            for instance in test_designspace.instances
            if instance.styleName in instance_names
        ]

    # Generate dummy UFOs for collected instances so we don't actually need to
    # interpolate.
    tmpdir.mkdir(instance_dir)
    for instance in test_instances:
        ufo = ufo_module.Font()
        ufo.save(str(tmpdir / instance))

    ufos = apply_instance_data(designspace.path, include_filenames=test_instances)

    for filename in test_instances:
        assert os.path.isdir(str(tmpdir / filename))
    assert len(ufos) == len(test_instances)

    for ufo in ufos:
        assert ufo.info.openTypeOS2WeightClass in {
            100,
            200,
            300,
            400,
            500,
            700,
            900,
            357,
        }
        assert ufo.info.openTypeOS2WidthClass is None  # GlyphsUnitTestSans is wght only


def test_reexport_apply_instance_data():
    # this is for compatibility with fontmake
    # https://github.com/googlefonts/fontmake/issues/451
    from glyphsLib.interpolation import apply_instance_data as reexported

    assert reexported is apply_instance_data


def test_reencode_glyphs(tmpdir):
    data_dir = py.path.local(DATA)

    designspace_path = data_dir / "TestReencode.designspace"
    designspace_path.copy(tmpdir)

    ufo_path = data_dir / "TestReencode-Regular.ufo"
    ufo_path.copy(tmpdir.ensure_dir("TestReencode-Regular.ufo"))

    instance_dir = tmpdir.ensure_dir("instance_ufo")
    ufo_path.copy(instance_dir.ensure_dir("TestReencode-Regular.ufo"))
    ufo_path.copy(instance_dir.ensure_dir("TestReencodeUI-Regular.ufo"))

    ufos = apply_instance_data(str(tmpdir / "TestReencode.designspace"))

    assert len(ufos) == 2
    assert ufos[0]["A"].unicode == 0x0041
    assert ufos[0]["A.alt"].unicode is None
    assert ufos[0]["C"].unicode == 0x0043
    # Reencode Glyphs: A.alt=0041, C=
    assert ufos[1]["A"].unicode is None
    assert ufos[1]["A.alt"].unicode == 0x0041
    assert ufos[1]["C"].unicode is None


def test_glyphs3_names():
    file = "InstanceFamilyName-G3.glyphs"
    font = glyphsLib.GSFont(os.path.join(DATA, file))

    expected_names = {
        "familyName": [
            "MyFamily",
            "MyFamily",
            "MyFamily 12pt",
            "MyFamily 12pt",
            "MyFamily 72pt",
            "MyFamily 72pt",
        ],
        "preferredFamily": [
            "MyFamily",
            "MyFamily",
            "MyFamily",
            "Typographic MyFamily 12pt",
            "MyFamily",
            "MyFamily",
        ],
        "preferredFamilyName": [
            None,
            None,
            None,
            "Typographic MyFamily 12pt",
            None,
            None,
        ],
        "preferredSubfamilyName": [
            None,
            None,
            None,
            None,
            None,
            "Typographic Black",
        ],
        "windowsFamily": [
            "MyFamily Thin",
            "MyFamily Black",
            "MyFamily 12pt Thin",
            "MyFamily 12pt Black",
            "MyFamily 72pt Thin",
            "MyFamily 72pt Black",
        ],
        "fontName": [
            "MyFamily-Thin",
            "MyFamily-Black",
            "MyFamily12pt-Thin",
            "MyFamily12pt-Black",
            "MyFamily72pt-Thin",
            "MyFamily72pt-Black",
        ],
        "fullName": [
            "MyFamily Thin",
            "MyFamily Black",
            "MyFamily 12pt Thin",
            "MyFamily 12pt Black",
            "MyFamily 72pt Thin",
            "MyFamily 72pt Black",
        ],
    }

    for name, expected in expected_names.items():
        actual = [getattr(instance, name) for instance in font.instances]
        assert expected == actual, name


def test_glyphs3_mapping():
    font = glyphsLib.GSFont(os.path.join(DATA, "Glyphs3Instances.glyphs"))
    # Instance1: designspace 200 -> userspace 400
    # Instance2: designspace 800 -> userspace 900
    # Instance2: designspace 600 -> userspace 650
    doc = glyphsLib.to_designspace(font)
    assert doc.axes[0].map == [(400, 200), (600, 650), (900, 800)]
    assert doc.instances[0].location == {"Weight": 200}
    assert doc.instances[1].location == {"Weight": 800}
    assert doc.instances[2].location == {"Weight": 650}


def test_glyphs3_instance_filtering():
    font = glyphsLib.GSFont(os.path.join(DATA, "InstanceFamilyName-G3.glyphs"))
    assert len(font.instances) == 6

    # Loaded from default font family name
    assert not font.instances[0].properties
    assert not font.instances[1].properties
    assert font.instances[0].familyName == "MyFamily"
    assert font.instances[1].familyName == "MyFamily"

    # Loaded from .properties
    assert font.instances[2].familyName == "MyFamily 12pt"
    assert font.instances[3].familyName == "MyFamily 12pt"
    assert font.instances[4].familyName == "MyFamily 72pt"
    assert font.instances[5].familyName == "MyFamily 72pt"

    doc = glyphsLib.to_designspace(font)
    assert len(doc.instances) == 6

    doc = glyphsLib.to_designspace(font, family_name="MyFamily 12pt")
    assert len(doc.instances) == 2


def test_glyphs3_instance_properties(tmpdir):
    expected_num_properties = [0, 0, 1, 2, 1, 2]

    file = "InstanceFamilyName-G3.glyphs"
    font = glyphsLib.GSFont(os.path.join(DATA, file))

    for expected, instance in zip(expected_num_properties, font.instances):
        assert expected == len(instance.properties)

    font.save(tmpdir / file)
    font = glyphsLib.GSFont(tmpdir / file)

    for expected, instance in zip(expected_num_properties, font.instances):
        assert expected == len(instance.properties)


def test_rename_glyphs(tmpdir):
    font = glyphsLib.GSFont(os.path.join(DATA, "RenameGlyphsTest.glyphs"))
    instance_dir = tmpdir.ensure_dir("instance_ufo")
    designspace = glyphsLib.to_designspace(font, instance_dir=instance_dir)
    path = str(tmpdir / (font.familyName + ".designspace"))
    write_designspace_and_UFOs(designspace, path)

    ufo_path = tmpdir / "RenameGlyphsTest-Regular.ufo"
    ufo_path.copy(instance_dir.ensure_dir("RenameGlyphsTest-Straight.ufo"))
    ufo_path.copy(instance_dir.ensure_dir("RenameGlyphsTest-Swapped.ufo"))

    ufos = apply_instance_data(designspace.path)

    assert len(ufos) == 2

    assert len(ufos[0]["a"][0]) == 4  # Square
    assert len(ufos[0]["b"][0]) == 12  # Circle
    assert ufos[0]["a"].unicode == 0x0061
    assert ufos[0]["b"].unicode == 0x0062

    assert len(ufos[1]["a"][0]) == 12  # Circle
    assert len(ufos[1]["b"][0]) == 4  # Square
    assert ufos[0]["a"].unicode == 0x0061
    assert ufos[0]["b"].unicode == 0x0062



# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os

from .axes import font_uses_axis_locations, get_axis_definitions
from .constants import (
    GLYPHS_PREFIX,
    MASTER_ID_LIB_KEY,
    UFO_YEAR_KEY,
    UFO_NOTE_KEY,
    UFO_FILENAME_CUSTOM_PARAM,
)


def to_ufo_master_attributes(self, ufo, master):
    ufo.info.ascender = master.ascender
    ufo.info.capHeight = master.capHeight
    ufo.info.descender = master.descender
    ufo.info.xHeight = master.xHeight

    horizontal_stems = master.horizontalStems
    vertical_stems = master.verticalStems
    italic_angle = -master.italicAngle
    if horizontal_stems:
        ufo.info.postscriptStemSnapH = horizontal_stems
    if vertical_stems:
        ufo.info.postscriptStemSnapV = vertical_stems
    if italic_angle is not None:
        ufo.info.italicAngle = italic_angle

    year = master.userData[UFO_YEAR_KEY]
    if year is not None:
        ufo.info.year = year
    note = master.userData[UFO_NOTE_KEY]
    if note is not None:
        ufo.info.note = note

    # All of this will go into the designspace as well
    # "Native" designspace fonts will only have the designspace info
    # FIXME: (jany) maybe we should not duplicate the information and only
    # write it in the designspace?
    widthValue = master.widthValue
    weightValue = master.weightValue
    if weightValue is not None:
        ufo.lib[GLYPHS_PREFIX + "weightValue"] = weightValue
    if widthValue:
        ufo.lib[GLYPHS_PREFIX + "widthValue"] = widthValue
    for number in ("", "1", "2", "3"):
        custom_value = getattr(master, "customValue" + number)
        if custom_value:
            ufo.lib[GLYPHS_PREFIX + "customValue" + number] = custom_value

    if font_uses_axis_locations(self.font):
        # Set the OS/2 weightClass and widthClas according the this master's
        # user location ("Axis Location" parameter)
        for axis in get_axis_definitions(self.font):
            if axis.tag in ("wght", "wdth"):
                user_loc = axis.get_user_loc(master)
                axis.set_ufo_user_loc(ufo, user_loc)

    # Set vhea values to glyphsapp defaults if they haven't been declared.
    # ufo2ft needs these set in order for a ufo to be recognised as
    # vertical. Glyphsapp uses the font upm, not the typo metrics
    # for these.
    if self.is_vertical:
        custom_params = master.customParameters
        font_upm = self.font.upm
        if not any(
            k in custom_params for k in ("vheaVertAscender", "vheaVertTypoAscender")
        ):
            ufo.info.openTypeVheaVertTypoAscender = int(font_upm / 2)
        if not any(
            k in custom_params for k in ("vheaVertDescender", "vheaVertTypoDescender")
        ):
            ufo.info.openTypeVheaVertTypoDescender = -int(font_upm / 2)
        if not any(
            k in custom_params for k in ("vheaVertLineGap", "vheaVertTypoLineGap")
        ):
            ufo.info.openTypeVheaVertTypoLineGap = font_upm

    self.to_ufo_blue_values(ufo, master)
    self.to_ufo_guidelines(ufo, master)
    self.to_ufo_master_user_data(ufo, master)
    # Note: master's custom parameters will be applied later on, after glyphs and
    # features have been generated (see UFOBuilder::masters method).

    master_id = master.id
    if self.minimize_glyphs_diffs:
        ufo.lib[MASTER_ID_LIB_KEY] = master_id


def to_glyphs_master_attributes(self, source, master):
    ufo = source.font

    # Glyphs ensures that the master ID is unique by simply making up a new one when
    # finding a duplicate.
    ufo_master_id_lib_key = ufo.lib.get(MASTER_ID_LIB_KEY)
    if ufo_master_id_lib_key and not self.font.masters[ufo_master_id_lib_key]:
        master.id = ufo_master_id_lib_key

    if source.filename is not None and self.minimize_ufo_diffs:
        master.customParameters[UFO_FILENAME_CUSTOM_PARAM] = source.filename
    elif ufo.path and self.minimize_ufo_diffs:
        # Don't be smart, we don't know where the UFOs come from so we can't make them
        # relative to anything.
        master.customParameters[UFO_FILENAME_CUSTOM_PARAM] = os.path.basename(ufo.path)

    if ufo.info.ascender is not None:
        master.ascender = ufo.info.ascender
    if ufo.info.capHeight is not None:
        master.capHeight = ufo.info.capHeight
    if ufo.info.descender is not None:
        master.descender = ufo.info.descender
    if ufo.info.xHeight is not None:
        master.xHeight = ufo.info.xHeight

    horizontal_stems = ufo.info.postscriptStemSnapH
    vertical_stems = ufo.info.postscriptStemSnapV
    italic_angle = 0
    if ufo.info.italicAngle:
        italic_angle = -ufo.info.italicAngle
    if horizontal_stems:
        master.horizontalStems = horizontal_stems
    if vertical_stems:
        master.verticalStems = vertical_stems
    if italic_angle:
        master.italicAngle = italic_angle

    if ufo.info.year is not None:
        master.userData[UFO_YEAR_KEY] = ufo.info.year
    if ufo.info.note is not None:
        master.userData[UFO_NOTE_KEY] = ufo.info.note

    self.to_glyphs_blue_values(ufo, master)
    self.to_glyphs_master_names(ufo, master)
    self.to_glyphs_master_user_data(ufo, master)
    self.to_glyphs_guidelines(ufo, master)
    self.to_glyphs_custom_params(ufo, master)



# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from io import open
import collections
import os
import logging

from glyphsLib.classes import GSFont, __all__ as __all_classes__
from glyphsLib.classes import *  # noqa
from glyphsLib.builder import to_ufos, to_designspace, to_glyphs  # noqa
from glyphsLib.parser import load, loads  # noqa
from glyphsLib.writer import dump, dumps  # noqa
from glyphsLib.util import clean_ufo, ufo_create_background_layer_for_all_glyphs

try:
    from ._version import version as __version__
except ImportError:
    __version__ = "0.0.0+unknown"

__all__ = [
    "build_masters",
    "load_to_ufos",
    "to_ufos",
    "to_designspace",
    "to_glyphs",
    "load",
    "loads",
    "dump",
    "dumps",
] + __all_classes__

logger = logging.getLogger(__name__)

Masters = collections.namedtuple("Masters", ["ufos", "designspace_path"])


def load_to_ufos(
    file_or_path,
    include_instances=False,
    family_name=None,
    propagate_anchors=None,
    ufo_module=None,
    expand_includes=False,
    minimal=False,
    glyph_data=None,
):
    """Load an unpacked .glyphs object to UFO objects."""

    if hasattr(file_or_path, "read"):
        font = load(file_or_path)
    else:
        with open(file_or_path, "r", encoding="utf-8") as ifile:
            font = load(ifile)
    logger.info("Loading to UFOs")
    return to_ufos(
        font,
        include_instances=include_instances,
        family_name=family_name,
        propagate_anchors=propagate_anchors,
        ufo_module=ufo_module,
        expand_includes=expand_includes,
        minimal=minimal,
        glyph_data=glyph_data,
    )


def build_masters(
    filename: str | bytes | os.PathLike[str] | os.PathLike[bytes] | GSFont,
    master_dir,
    designspace_instance_dir=None,
    designspace_path=None,
    family_name=None,
    propagate_anchors=None,
    minimize_glyphs_diffs=False,
    normalize_ufos=False,
    create_background_layers=False,
    generate_GDEF=True,
    store_editor_state=True,
    write_skipexportglyphs=False,
    expand_includes=False,
    ufo_module=None,
    minimal=False,
    glyph_data=None,
):
    """Write and return UFOs from the masters and the designspace defined in a
    .glyphs file.

    Args:
        filename: Path to Glyphs sources, or GSFont object (may be mutated)
        master_dir: Directory where masters are written.
        designspace_instance_dir: If provided, a designspace document will be
            written alongside the master UFOs though no instances will be built.
        family_name: If provided, the master UFOs will be given this name and
            only instances with this name will be included in the designspace.

    Returns:
        A named tuple of master UFOs (`ufos`) and the path to the designspace
        file (`designspace_path`).
    """

    # Either use 'filename' as a loaded font, or interpret it as a path.
    # (variable name is 'filename' in both cases for backwards compatibility)
    if isinstance(filename, GSFont):
        font = filename
    else:
        font = GSFont(filename)

    if not os.path.isdir(master_dir):
        os.mkdir(master_dir)

    if designspace_instance_dir is None:
        instance_dir = None
    else:
        instance_dir = os.path.relpath(designspace_instance_dir, master_dir)

    designspace = to_designspace(
        font,
        family_name=family_name,
        propagate_anchors=propagate_anchors,
        instance_dir=instance_dir,
        minimize_glyphs_diffs=minimize_glyphs_diffs,
        generate_GDEF=generate_GDEF,
        store_editor_state=store_editor_state,
        write_skipexportglyphs=write_skipexportglyphs,
        expand_includes=expand_includes,
        ufo_module=ufo_module,
        minimal=minimal,
        glyph_data=glyph_data,
    )

    # Only write full masters to disk. This assumes that layer sources are always part
    # of another full master source, which must always be the case in a .glyphs file.
    ufos = {}
    for source in designspace.sources:
        if source.filename in ufos:
            assert source.font is ufos[source.filename]
            continue

        if create_background_layers and not minimal:
            ufo_create_background_layer_for_all_glyphs(source.font)

        ufo_path = os.path.join(master_dir, source.filename)
        clean_ufo(ufo_path)
        source.font.save(ufo_path)

        if normalize_ufos:
            import ufonormalizer

            ufonormalizer.normalizeUFO(ufo_path, writeModTimes=False)

        ufos[source.filename] = source.font

    if not designspace_path:
        designspace_path = os.path.join(master_dir, designspace.filename)
    designspace.write(designspace_path)

    return Masters(ufos, designspace_path)



# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from glyphsLib.util import pairs


def to_ufo_blue_values(self, ufo, master):
    """Set postscript blue values from Glyphs alignment zones."""

    alignment_zones = master.alignmentZones
    blue_values = []
    other_blues = []
    for zone in sorted(alignment_zones):
        pos = zone.position
        size = zone.size
        val_list = blue_values if pos == 0 or size >= 0 else other_blues
        val_list.extend(sorted((pos, pos + size)))

    if blue_values:
        ufo.info.postscriptBlueValues = blue_values
    if other_blues:
        ufo.info.postscriptOtherBlues = other_blues


def to_glyphs_blue_values(self, ufo, master):
    """Sets the GSFontMaster alignmentZones from the postscript blue values."""

    zones = []
    blue_values = pairs(ufo.info.postscriptBlueValues or [])
    other_blues = pairs(ufo.info.postscriptOtherBlues or [])
    for y1, y2 in blue_values:
        size = y2 - y1
        if y2 == 0:
            pos = 0
            size = -size
        else:
            pos = y1
        zones.append(self.glyphs_module.GSAlignmentZone(pos, size))
    for y1, y2 in other_blues:
        size = y1 - y2
        pos = y2
        zones.append(self.glyphs_module.GSAlignmentZone(pos, size))

    master.alignmentZones = sorted(zones, key=lambda zone: -zone.position)



# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# TODO: (jany) merge with builder/common.py

import logging
import itertools
import os
import shutil
from fontTools.misc.textTools import num2binary

logger = logging.getLogger(__name__)


def build_ufo_path(out_dir, family_name, style_name):
    """Build string to use as a UFO path."""

    return os.path.join(
        out_dir,
        "%s-%s.ufo"
        % ((family_name or "").replace(" ", ""), (style_name or "").replace(" ", "")),
    )


def open_ufo(path, font_class, **kwargs):
    try:
        return font_class.open(path, lazy=False, **kwargs)  # ufoLib2
    except AttributeError:
        return font_class(path, **kwargs)  # defcon, fontParts, etc.


def write_ufo(ufo, out_dir):
    """Write a UFO."""

    out_path = build_ufo_path(out_dir, ufo.info.familyName, ufo.info.styleName)

    logger.info("Writing %s" % out_path)
    clean_ufo(out_path)
    ufo.save(out_path)


def clean_ufo(path):
    """Make sure old UFO data is removed, as it may contain deleted glyphs."""

    if path.endswith(".ufo") and os.path.exists(path):
        shutil.rmtree(path)


def ufo_create_background_layer_for_all_glyphs(ufo_font):
    """Create a background layer for all glyphs in ufo_font if not present to
    reduce roundtrip differences."""

    if "public.background" in ufo_font.layers:
        background = ufo_font.layers["public.background"]
    else:
        background = ufo_font.newLayer("public.background")

    for glyph in ufo_font:
        if glyph.name not in background:
            background.newGlyph(glyph.name)


def cast_to_number_or_bool(inputstr):
    """Cast a string to int, float or bool. Return original string if it can't be
    converted.

    Scientific expression is converted into float.
    """
    if inputstr.strip().lower() == "true":
        return True
    elif inputstr.strip().lower() == "false":
        return False
    try:
        return int(inputstr)
    except ValueError:
        try:
            return float(inputstr)
        except ValueError:
            return inputstr


def reverse_cast_to_number_or_bool(input):
    if input is True:
        return "true"  # FIXME: (jany) dubious, glyphs handbook says should be 1
    if input is False:
        return "false"  # FIXME: (jany) dubious, glyphs handbook says should be 0
    return str(input)


def bin_to_int_list(value):
    string = num2binary(value)
    string = string.replace(" ", "")  # num2binary add a space every 8 digits
    return [i for i, v in enumerate(reversed(string)) if v == "1"]


def int_list_to_bin(value):
    result = 0
    for i in value:
        result += 1 << i
    return result


def pairwise(iterable):
    "s -> (s0,s1), (s1,s2), (s2, s3), ..."
    a, b = itertools.tee(iterable)
    next(b, None)
    return zip(a, b)


def tostr(s, encoding="ascii", errors="strict"):
    if not isinstance(s, str):
        return s.decode(encoding, errors)
    else:
        return s


def pairs(list):
    "s -> (s0,s1), (s2,s3), (s4, s5), ..."
    return [list[i : i + 2] for i in range(0, len(list), 2)]


def freezedict(dct):
    return frozenset(dct.items())


class LoggerMixin:
    _logger = None

    @property
    def logger(self):
        if self._logger is None:
            self._logger = logging.getLogger(
                ".".join([self.__class__.__module__, self.__class__.__name__])
            )
        return self._logger


def designspace_min_max(axis):
    """Return the minimum/maximum of an axis in designspace coordinates"""
    if not axis.map:
        return axis.minimum, axis.maximum
    designspace_scale = [design_location for _, design_location in axis.map]
    return min(designspace_scale), max(designspace_scale)


class PeekableIterator:
    """Helper class to iterate and peek over a list."""

    def __init__(self, list):
        self.index = 0
        self.list = list

    def has_next(self, n=0):
        return (self.index + n) < len(self.list)

    def __iter__(self):
        return self

    def __next__(self):
        res = self.list[self.index]
        self.index += 1
        return res

    next = __next__

    def peek(self, n=0):
        return self.list[self.index + n]


# sentinel object to indicate a deprecated argument
_DeprecatedArgument = object()



#
# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import copy
import os
import unittest
from textwrap import dedent

from unittest import mock
from unittest.mock import patch

import glyphsLib
import defcon
import ufoLib2
from glyphsLib.builder.builders import UFOBuilder
from glyphsLib.builder import to_ufos
from glyphsLib.builder.custom_params import (
    _set_default_params,
    GLYPHS_UFO_CUSTOM_PARAMS,
)
from glyphsLib.builder.constants import (
    UFO2FT_FILTERS_KEY,
    UFO2FT_USE_PROD_NAMES_KEY,
    FONT_CUSTOM_PARAM_PREFIX,
    MASTER_CUSTOM_PARAM_PREFIX,
    UFO_FILENAME_CUSTOM_PARAM,
    GLYPHLIB_PREFIX,
    UFO_FILENAME_KEY,
    FULL_FILENAME_KEY,
)
from glyphsLib.classes import GSFont, GSFontMaster, GSCustomParameter, GSGlyph, GSLayer
from glyphsLib.types import parse_datetime

import pytest


DATA = os.path.join(os.path.dirname(os.path.dirname(__file__)), "data")


class SetCustomParamsTestBase(object):
    ufo_module = None  # subclasses must override this

    def setUp(self):
        self.ufo = self.ufo_module.Font()
        self.font = GSFont()
        self.master = GSFontMaster()
        self.font.masters.insert(0, self.master)
        self.builder = UFOBuilder(self.font)

    def set_custom_params(self):
        self.builder.to_ufo_custom_params(self.ufo, self.font)
        self.builder.to_ufo_custom_params(self.ufo, self.master)

    def test_normalizes_curved_quotes_in_names(self):
        self.master.customParameters = [
            GSCustomParameter(name="‘bad’", value=1),
            GSCustomParameter(name="“also bad”", value=2),
        ]
        self.set_custom_params()
        self.assertIn(MASTER_CUSTOM_PARAM_PREFIX + "'bad'", self.ufo.lib)
        self.assertIn(MASTER_CUSTOM_PARAM_PREFIX + '"also bad"', self.ufo.lib)

    def test_set_fsSelection_flags_none(self):
        self.ufo.info.openTypeOS2Selection = None
        self.font = glyphsLib.to_glyphs([self.ufo], minimize_ufo_diffs=True)
        self.assertEqual(self.font.customParameters["Use Typo Metrics"], None)
        self.assertEqual(self.font.customParameters["Has WWS Names"], None)
        self.assertEqual(
            self.font.customParameters["openTypeOS2SelectionUnsupportedBits"], None
        )
        self.set_custom_params()
        self.assertEqual(self.ufo.info.openTypeOS2Selection, None)

    def test_set_fsSelection_flags_empty(self):
        self.ufo.info.openTypeOS2Selection = []
        self.font = glyphsLib.to_glyphs([self.ufo], minimize_ufo_diffs=True)
        self.assertEqual(self.font.customParameters["Use Typo Metrics"], None)
        self.assertEqual(self.font.customParameters["Has WWS Names"], None)
        self.assertEqual(
            self.font.customParameters["openTypeOS2SelectionUnsupportedBits"], []
        )
        self.set_custom_params()
        self.assertEqual(self.ufo.info.openTypeOS2Selection, [])

    def test_set_fsSelection_flags_all(self):
        self.ufo.info.openTypeOS2Selection = [1, 2, 3, 4, 7, 8, 9]
        self.font = glyphsLib.to_glyphs([self.ufo], minimize_ufo_diffs=True)
        self.assertEqual(self.font.customParameters["Use Typo Metrics"], True)
        self.assertEqual(self.font.customParameters["Has WWS Names"], True)
        self.assertEqual(
            self.font.customParameters["openTypeOS2SelectionUnsupportedBits"],
            [1, 2, 3, 4, 9],
        )
        self.set_custom_params()
        self.assertEqual(self.ufo.info.openTypeOS2Selection, [1, 2, 3, 4, 7, 8, 9])

    def test_set_fsSelection_flags(self):
        self.assertEqual(self.ufo.info.openTypeOS2Selection, None)

        self.master.customParameters["Has WWS Names"] = False
        self.set_custom_params()
        self.assertEqual(self.ufo.info.openTypeOS2Selection, None)

        self.master.customParameters["Use Typo Metrics"] = True
        self.set_custom_params()
        self.assertEqual(self.ufo.info.openTypeOS2Selection, [7])

        self.ufo = self.ufo_module.Font()
        self.master.customParameters = [
            GSCustomParameter(name="Use Typo Metrics", value=True),
            GSCustomParameter(name="Has WWS Names", value=True),
        ]
        self.set_custom_params()
        self.assertEqual(self.ufo.info.openTypeOS2Selection, [7, 8])

    def test_integer_parameters(self):
        """Test casting glyphsapp customParameters whose values are just
        integers into ufo equivalents."""
        integer_params = [
            "underlinePosition",
            "underlineThickness",
            "strikeoutPosition",
            "strikeoutSize",
            "subscriptXSize",
            "subscriptYSize",
            "subscriptXOffset",
            "subscriptYOffset",
            "superscriptXSize",
            "superscriptYSize",
            "superscriptXOffset",
            "superscriptYOffset",
        ]
        params_to_check = [
            (k, v) for (k, v) in GLYPHS_UFO_CUSTOM_PARAMS if k in integer_params
        ]

        for glyphs_key, ufo_key in params_to_check:
            self.master.customParameters[glyphs_key] = 10
            self.set_custom_params()
            self.assertEqual(getattr(self.ufo.info, ufo_key), 10)
            for param in self.master.customParameters:
                if param.name == glyphs_key:
                    param.value = -2
                    break
            self.set_custom_params()
            self.assertEqual(getattr(self.ufo.info, ufo_key), -2)

    @patch("glyphsLib.builder.custom_params.parse_glyphs_filter")
    def test_parse_glyphs_filter(self, mock_parse_glyphs_filter):
        pre_filter = "AddExtremes"
        filter1 = "Transformations;OffsetX:40;OffsetY:60;include:uni0334,uni0335"
        filter2 = "Transformations;OffsetX:10;OffsetY:-10;exclude:uni0334,uni0335"
        self.master.customParameters.extend(
            [
                GSCustomParameter(name="PreFilter", value=pre_filter),
                GSCustomParameter(name="Filter", value=filter1),
                GSCustomParameter(name="Filter", value=filter2),
            ]
        )
        self.set_custom_params()

        self.assertEqual(mock_parse_glyphs_filter.call_count, 3)
        self.assertEqual(
            mock_parse_glyphs_filter.call_args_list[0],
            mock.call(pre_filter, is_pre=True),
        )
        self.assertEqual(
            mock_parse_glyphs_filter.call_args_list[1], mock.call(filter1, is_pre=False)
        )
        self.assertEqual(
            mock_parse_glyphs_filter.call_args_list[2], mock.call(filter2, is_pre=False)
        )

    def test_set_defaults(self):
        _set_default_params(self.ufo)
        self.assertEqual(self.ufo.info.openTypeOS2Type, [3])
        self.assertEqual(self.ufo.info.postscriptUnderlinePosition, -100)
        self.assertEqual(self.ufo.info.postscriptUnderlineThickness, 50)

    def test_set_codePageRanges_empty(self):
        self.font.customParameters["codePageRanges"] = []
        self.set_custom_params()
        self.assertEqual(self.ufo.info.openTypeOS2CodePageRanges, [])
        self.font = glyphsLib.to_glyphs([self.ufo], minimize_ufo_diffs=True)
        self.assertEqual(self.font.customParameters["codePageRanges"], [])

    def test_set_codePageRanges(self):
        self.font.customParameters["codePageRanges"] = [1252, 1250]
        self.font.customParameters["codePageRangesUnsupportedBits"] = [15]
        self.set_custom_params()
        self.assertEqual(self.ufo.info.openTypeOS2CodePageRanges, [0, 1, 15])
        self.font = glyphsLib.to_glyphs([self.ufo], minimize_ufo_diffs=True)
        self.assertEqual(self.font.customParameters["codePageRanges"], [1252, 1250])
        self.assertEqual(
            self.font.customParameters["codePageRangesUnsupportedBits"], [15]
        )

    def test_set_codePageRanges_with_various_types(self):
        self.font.customParameters["openTypeOS2CodePageRanges"] = [
            1252,
            "1250",
            "bit 29",
        ]
        self.font.customParameters["codePageRangesUnsupportedBits"] = [15]
        self.set_custom_params()
        self.assertEqual(self.ufo.info.openTypeOS2CodePageRanges, [0, 1, 15, 29])
        self.font = glyphsLib.to_glyphs([self.ufo], minimize_ufo_diffs=True)
        self.assertEqual(self.font.customParameters["codePageRanges"], [1252, 1250])
        self.assertEqual(
            self.font.customParameters["codePageRangesUnsupportedBits"], [15, 29]
        )

    def test_set_openTypeOS2CodePageRanges(self):
        self.font.customParameters["openTypeOS2CodePageRanges"] = [1252, 1250]
        self.font.customParameters["codePageRangesUnsupportedBits"] = [15]
        self.set_custom_params()
        self.assertEqual(self.ufo.info.openTypeOS2CodePageRanges, [0, 1, 15])
        self.font = glyphsLib.to_glyphs([self.ufo], minimize_ufo_diffs=True)
        self.assertEqual(self.font.customParameters["codePageRanges"], [1252, 1250])
        self.assertEqual(
            self.font.customParameters["codePageRangesUnsupportedBits"], [15]
        )

    def test_gasp_table(self):
        gasp_table = {"65535": "15", "20": "7", "8": "10"}
        self.font.customParameters["GASP Table"] = gasp_table
        self.set_custom_params()

        ufo_range_records = self.ufo.info.openTypeGaspRangeRecords
        self.assertIsNotNone(ufo_range_records)
        self.assertEqual(len(ufo_range_records), 3)
        rec1, rec2, rec3 = ufo_range_records
        self.assertEqual(rec1["rangeMaxPPEM"], 8)
        self.assertEqual(rec1["rangeGaspBehavior"], [1, 3])
        self.assertEqual(rec2["rangeMaxPPEM"], 20)
        self.assertEqual(rec2["rangeGaspBehavior"], [0, 1, 2])
        self.assertEqual(rec3["rangeMaxPPEM"], 65535)
        self.assertEqual(rec3["rangeGaspBehavior"], [0, 1, 2, 3])

    def test_set_disables_nice_names(self):
        self.font.disablesNiceNames = False
        self.set_custom_params()
        self.assertEqual(True, self.ufo.lib[FONT_CUSTOM_PARAM_PREFIX + "useNiceNames"])

    def test_set_disable_last_change(self):
        glyph = GSGlyph()
        glyph.name = "a"
        self.font.glyphs.append(glyph)
        layer = GSLayer()
        layer.layerId = self.font.masters[0].id
        layer.associatedMasterId = self.font.masters[0].id
        layer.width = 100
        glyph.layers.append(layer)
        glyph.lastChange = parse_datetime("2017-10-03 07:35:46 +0000")
        self.font.customParameters["Disable Last Change"] = True
        self.ufo = to_ufos(self.font)[0]
        self.assertEqual(
            True, self.ufo.lib[FONT_CUSTOM_PARAM_PREFIX + "disablesLastChange"]
        )
        self.assertNotIn(GLYPHLIB_PREFIX + "lastChange", self.ufo["a"].lib)

    # https://github.com/googlefonts/glyphsLib/issues/268
    def test_xHeight(self):
        self.ufo.info.xHeight = 300
        self.master.customParameters["xHeight"] = "500"
        self.set_custom_params()
        # Additional xHeight values are Glyphs-specific and stored in lib
        self.assertEqual(self.ufo.lib[MASTER_CUSTOM_PARAM_PREFIX + "xHeight"], "500")
        # The xHeight from the property is not modified
        self.assertEqual(self.ufo.info.xHeight, 300)
        # TODO: (jany) check that the instance custom param wins over the
        #       interpolated value

    def test_replace_feature(self):
        self.ufo.features.text = dedent(
            """
            feature liga {
            # only the first match is replaced
            sub f i by fi;
            } liga;

            feature calt {
            sub e' t' c by ampersand;
            } calt;

            feature liga {
            sub f l by fl;
            } liga;
        """
        )

        repl = "liga; sub f f by ff;"

        self.master.customParameters["Replace Feature"] = repl
        self.set_custom_params()

        self.assertEqual(
            self.ufo.features.text,
            dedent(
                """
            feature liga {
            sub f f by ff;
            } liga;

            feature calt {
            sub e' t' c by ampersand;
            } calt;

            feature liga {
            sub f l by fl;
            } liga;
        """
            ),
        )

        # only replace feature body if tag already present
        original = self.ufo.features.text
        repl = "numr; sub one by one.numr;\nsub two by two.numr;\n"

        self.master.customParameters["Replace Feature"] = repl
        self.set_custom_params()

        self.assertEqual(self.ufo.features.text, original)

    def test_replace_prefix(self):
        self.ufo.features.text = dedent(
            """\
            # Prefix: AAA
            include(../aaa.fea);

            # Prefix: FOO
            # foo

            # Prefix: ZZZ
            include(../zzz.fea);

            # Prefix: BAR
            # bar

            feature liga {
            sub f i by f_i;
            } liga;

            table GDEF {
            GlyphClassDef
                [f i], # Base
                [f_i], # Liga
                , # Mark
                ;
            } GDEF;
            """
        )

        self.master.customParameters.append(
            GSCustomParameter("Replace Prefix", "FOO; include(../foo.fea);")
        )
        self.master.customParameters.append(
            GSCustomParameter("Replace Prefix", "BAR; include(../bar.fea);")
        )
        self.set_custom_params()

        self.assertEqual(
            self.ufo.features.text,
            dedent(
                """\
                # Prefix: AAA
                include(../aaa.fea);

                # Prefix: FOO
                include(../foo.fea);

                # Prefix: ZZZ
                include(../zzz.fea);

                # Prefix: BAR
                include(../bar.fea);

                table GDEF {
                GlyphClassDef
                    [f i], # Base
                    [f_i], # Liga
                    , # Mark
                    ;
                } GDEF;

                feature liga {
                sub f i by f_i;
                } liga;
                """
            ),
        )

    def test_useProductionNames(self):
        for value in (True, False):
            self.master.customParameters["Don't use Production Names"] = value
            self.set_custom_params()

            self.assertIn(UFO2FT_USE_PROD_NAMES_KEY, self.ufo.lib)
            self.assertEqual(self.ufo.lib[UFO2FT_USE_PROD_NAMES_KEY], not value)

    def test_default_fstype(self):
        # No specified fsType => set default value
        self.set_custom_params()
        self.assertEqual(self.ufo.info.openTypeOS2Type, [3])

    def test_set_fstype(self):
        # Set another fsType => store that
        self.master.customParameters["fsType"] = [2]
        self.set_custom_params()
        self.assertEqual(self.ufo.info.openTypeOS2Type, [2])

    def test_empty_fstype(self):
        # Set empty fsType => store empty
        self.master.customParameters["fsType"] = []
        self.set_custom_params()
        self.assertEqual(self.ufo.info.openTypeOS2Type, [])

    def test_version_string(self):
        # TODO: (jany) test the automatic replacement that is described in the
        #   Glyphs Handbook
        self.font.customParameters["versionString"] = "Version 2.040"
        self.set_custom_params()
        self.assertEqual(self.ufo.info.openTypeNameVersion, "Version 2.040")

    def test_ufo2ft_filter_roundtrip(self):
        ufo_filters = [
            {"name": "propagateAnchors", "pre": True, "include": ["a", "b", "c"]}
        ]
        glyphs_filter = "propagateAnchors;include:a,b,c"

        # Test the one-way conversion of (Pre)Filters into ufo2ft filters. See the
        # docstring for FilterParamHandler.
        self.master.customParameters["PreFilter"] = glyphs_filter
        self.set_custom_params()
        self.assertEqual(self.ufo.lib[UFO2FT_FILTERS_KEY], ufo_filters)

        # Test the round-tripping of ufo2ft filters from UFO -> Glyphs master -> UFO.
        # See the docstring for FilterParamHandler.
        font_rt = glyphsLib.to_glyphs([self.ufo])
        self.assertNotIn("PreFilter", font_rt.masters[0].customParameters)
        self.assertEqual(font_rt.masters[0].userData[UFO2FT_FILTERS_KEY], ufo_filters)
        ufo_rt = glyphsLib.to_ufos(font_rt, ufo_module=self.ufo_module)[0]
        self.assertEqual(ufo_rt.lib[UFO2FT_FILTERS_KEY], ufo_filters)

    def test_color_palettes(self):
        glyphs_palettes = [
            ["68,0,59,255", "220,187,72,255", "42,255", "87,255", "0,138,255,255"]
        ]
        ufo_palettes = [
            [
                (0.26666666666666666, 0.0, 0.23137254901960785, 1.0),
                (0.8627450980392157, 0.7333333333333333, 0.2823529411764706, 1.0),
                (0.16470588235294117, 0.16470588235294117, 0.16470588235294117, 1.0),
                (0.3411764705882353, 0.3411764705882353, 0.3411764705882353, 1.0),
                (0.0, 0.5411764705882353, 1.0, 1.0),
            ]
        ]
        self.font.customParameters["Color Palettes"] = glyphs_palettes
        self.set_custom_params()
        self.assertEqual(
            self.ufo.lib["com.github.googlei18n.ufo2ft.colorPalettes"], ufo_palettes
        )

        # Test the round-tripping
        font = glyphsLib.to_glyphs([self.ufo])
        self.assertEqual(font.customParameters["Color Palettes"], glyphs_palettes)

    def test_meta_table(self):
        glyphs_meta = [
            {"data": "de-Latn", "tag": "dlng"},
            {"data": "en-Latn", "tag": "dlng"},
            {"data": "sr-Cyrl", "tag": "slng"},
            {"data": "\x00\x00...", "tag": "appl"},
        ]
        self.font.customParameters["meta Table"] = glyphs_meta
        ufo_meta = {
            "dlng": ["de-Latn", "en-Latn"],
            "slng": ["sr-Cyrl"],
            "appl": "\x00\x00...",
        }
        self.set_custom_params()
        self.assertEqual(self.ufo.lib["public.openTypeMeta"], ufo_meta)

        font = glyphsLib.to_glyphs([self.ufo])
        self.assertEqual(font.customParameters["meta Table"], glyphs_meta)

    def test_name_unique_id_as_property(self):
        uniqueID = "Foo Bar: Version 1.234"
        self.font.properties["uniqueID"] = uniqueID
        self.set_custom_params()
        self.assertEqual(self.ufo.info.openTypeNameUniqueID, uniqueID)
        # we currently do not round-trip properties
        # font = glyphsLib.to_glyphs([self.ufo])
        # self.assertEqual(font.properties["uniqueID"], uniqueID)

    def test_name_unique_id_as_custom_parameter(self):
        uniqueID = "Foo Bar: Version 1.234"
        self.font.customParameters["uniqueID"] = uniqueID
        self.set_custom_params()
        self.assertEqual(self.ufo.info.openTypeNameUniqueID, uniqueID)
        font = glyphsLib.to_glyphs([self.ufo])
        self.assertEqual(font.customParameters["uniqueID"], uniqueID)

    def test_name_table_entry(self):
        self.font.customParameters.append(
            GSCustomParameter("Name Table Entry", "1024; FOO; BAZ")
        )
        self.font.customParameters.append(
            GSCustomParameter("Name Table Entry", "2048 1; FOO")
        )
        self.font.customParameters.append(
            GSCustomParameter("Name Table Entry", "4096 1 2; FOO")
        )
        self.font.customParameters.append(
            GSCustomParameter("Name Table Entry", "8192 1 2 3; FOO")
        )
        self.font.customParameters.append(
            GSCustomParameter("Name Table Entry", "0x4000 074; BAZ")
        )

        self.set_custom_params()

        ufo_records = [
            {
                "nameID": 1024,
                "platformID": 3,
                "encodingID": 1,
                "languageID": 0x409,
                "string": "FOO; BAZ",
            },
            {
                "nameID": 2048,
                "platformID": 1,
                "encodingID": 0,
                "languageID": 0,
                "string": "FOO",
            },
            {
                "nameID": 4096,
                "platformID": 1,
                "encodingID": 2,
                "languageID": 0,
                "string": "FOO",
            },
            {
                "nameID": 8192,
                "platformID": 1,
                "encodingID": 2,
                "languageID": 3,
                "string": "FOO",
            },
            {
                "nameID": 16384,
                "platformID": 60,
                "encodingID": 1,
                "languageID": 0x409,
                "string": "BAZ",
            },
        ]

        self.assertEqual(
            [dict(r) for r in self.ufo.info.openTypeNameRecords], ufo_records
        )

        font = glyphsLib.to_glyphs([self.ufo])

        self.assertEqual(font.customParameters[0].value, "1024 3 1 1033; FOO; BAZ")
        self.assertEqual(font.customParameters[1].value, "2048 1 0 0; FOO")
        self.assertEqual(font.customParameters[2].value, "4096 1 2 0; FOO")
        self.assertEqual(font.customParameters[3].value, "8192 1 2 3; FOO")
        self.assertEqual(font.customParameters[4].value, "16384 60 1 1033; BAZ")


class SetCustomParamsTestUfoLib2(SetCustomParamsTestBase, unittest.TestCase):
    ufo_module = ufoLib2


class SetCustomParamsTestDefcon(SetCustomParamsTestBase, unittest.TestCase):
    ufo_module = defcon


def test_ufo_filename(ufo_module):
    """Test that new-style UFO_FILENAME_CUSTOM_PARAM is written instead of
    (UFO_FILENAME_KEY|FULL_FILENAME_KEY)."""
    font = glyphsLib.GSFont(os.path.join(DATA, "UFOFilenameTest.glyphs"))
    ds = glyphsLib.to_designspace(
        font, minimize_glyphs_diffs=True, ufo_module=ufo_module
    )
    assert ds.sources[0].filename == "MyFontMaster.ufo"
    assert ds.instances[0].filename == "../build/instance_ufos/MyFont.ufo"
    assert "com.schriftgestaltung.customParameters" not in ds.instances[0].lib

    font_rt = glyphsLib.to_glyphs(ds, minimize_ufo_diffs=True)
    assert (
        font_rt.masters[0].customParameters[UFO_FILENAME_CUSTOM_PARAM]
        == "MyFontMaster.ufo"
    )
    assert UFO_FILENAME_KEY not in font_rt.masters[0].userData
    assert (
        font_rt.instances[0].customParameters[UFO_FILENAME_CUSTOM_PARAM]
        == "../build/instance_ufos/MyFont.ufo"
    )
    assert FULL_FILENAME_KEY not in font_rt.instances[0].customParameters

    ds_rt = glyphsLib.to_designspace(
        font_rt, minimize_glyphs_diffs=True, ufo_module=ufo_module
    )
    assert ds_rt.sources[0].filename == "MyFontMaster.ufo"
    assert ds_rt.instances[0].filename == "../build/instance_ufos/MyFont.ufo"


def test_ufo_filename_with_legacy(ufo_module):
    """Test that new-style UFO_FILENAME_CUSTOM_PARAM overrides legacy
    (UFO_FILENAME_KEY|FULL_FILENAME_KEY)."""
    font = glyphsLib.GSFont(os.path.join(DATA, "UFOFilenameTest.glyphs"))
    font.masters[0].customParameters[UFO_FILENAME_CUSTOM_PARAM] = "aaa.ufo"
    font.instances[0].customParameters[UFO_FILENAME_CUSTOM_PARAM] = "bbb.ufo"

    ds = glyphsLib.to_designspace(
        font, minimize_glyphs_diffs=True, ufo_module=ufo_module
    )
    assert ds.sources[0].filename == "aaa.ufo"
    assert ds.instances[0].filename == "bbb.ufo"


def test_ufo_filename_with_instance_empty(ufo_module):
    font = glyphsLib.GSFont(os.path.join(DATA, "UFOFilenameTest.glyphs"))
    font.masters[0].customParameters[UFO_FILENAME_CUSTOM_PARAM] = "aaa.ufo"
    del font.instances[0].customParameters[UFO_FILENAME_CUSTOM_PARAM]
    del font.instances[0].customParameters[FULL_FILENAME_KEY]

    ds = glyphsLib.to_designspace(
        font, minimize_glyphs_diffs=True, ufo_module=ufo_module
    )
    assert ds.sources[0].filename == "aaa.ufo"
    # Instance filename should be whatever the default is.
    assert ds.instances[0].filename == "instance_ufos/NewFont-Regular.ufo"


def test_ufo_opentype_name_preferred_family_subfamily_name():
    from glyphsLib.interpolation import apply_instance_data_to_ufo

    filenames = [
        "UFOInstanceParametersTestV2.glyphs",
        # NOTE: In the format of version 3, the preferred family and subfamily
        # names are not actually saved in custom paramters but properties.
        "UFOInstanceParametersTestV3.glyphs",
    ]

    for filename in filenames:
        file = glyphsLib.GSFont(os.path.join(DATA, filename))
        space = glyphsLib.to_designspace(file, minimal=True)

        assert len(space.sources) == 2, filename
        assert len(space.instances) == 3, filename
        for instance, name in zip(space.instances, ["Thin", "Regular", "Black"]):
            source = copy.deepcopy(space.sources[0])
            apply_instance_data_to_ufo(source.font, instance, space)

            actual = source.font.info.openTypeNamePreferredFamilyName
            assert actual == "Typographic New Font", filename

            actual = source.font.info.openTypeNamePreferredSubfamilyName
            assert actual == f"Typographic {name}", filename


def test_ufo_opentype_name_records():
    from glyphsLib.interpolation import apply_instance_data_to_ufo

    filenames = [
        "UFOInstanceParametersTestV2.glyphs",
        "UFOInstanceParametersTestV3.glyphs",
    ]

    for filename in filenames:
        file = glyphsLib.GSFont(os.path.join(DATA, filename))
        space = glyphsLib.to_designspace(file, minimal=True)

        assert len(space.sources) == 2, filename
        for source in space.sources:
            actual = list(map(dict, source.font.info.openTypeNameRecords))
            expected = [
                {
                    "nameID": 42,
                    "platformID": 0,
                    "encodingID": 4,
                    "languageID": 0,
                    "string": "File",
                },
            ]
            assert actual == expected, filename

        assert len(space.instances) == 3, filename
        for instance, name in zip(space.instances, ["Thin", "Regular", "Black"]):
            source = copy.deepcopy(space.sources[0])
            apply_instance_data_to_ufo(source.font, instance, space)

            actual = list(map(dict, source.font.info.openTypeNameRecords))
            expected = [
                {
                    "nameID": 42,
                    "platformID": 0,
                    "encodingID": 4,
                    "languageID": 0,
                    "string": "File",
                },
                {
                    "nameID": 43,
                    "platformID": 0,
                    "encodingID": 4,
                    "languageID": 0,
                    "string": f"{name} Instance",
                },
            ]
            assert actual == expected, filename


def test_ufo_opentype_os2_selection():
    from glyphsLib.interpolation import apply_instance_data_to_ufo

    filenames = [
        "UFOInstanceParametersTestV2.glyphs",
        "UFOInstanceParametersTestV3.glyphs",
    ]

    for filename in filenames:
        file = glyphsLib.GSFont(os.path.join(DATA, filename))
        space = glyphsLib.to_designspace(file, minimal=True)

        assert len(space.sources) == 2, filename
        assert len(space.instances) == 3, filename
        for instance in space.instances:
            source = copy.deepcopy(space.sources[0])
            apply_instance_data_to_ufo(source.font, instance, space)

            actual = source.font.info.openTypeOS2Selection
            assert actual == [7, 8], filename


def test_mutiple_params(ufo_module):
    """Test multiple custom parameters with the same name on GSFont."""

    font = GSFont(os.path.join(DATA, "CustomPrametersTest.glyphs"))
    assert len(font.customParameters) == 3

    assert all("Virtual Master" == c.name for c in font.customParameters)

    assert font.customParameters[0].value == [{"Axis": "Spacing", "Location": 0}]
    assert font.customParameters[1].value == [{"Axis": "Spacing", "Location": -100}]
    assert font.customParameters[2].value == [{"Axis": "Spacing", "Location": 100}]

    instance = font.instances[0]
    assert len(instance.customParameters) == 2

    assert all("Replace Feature" == c.name for c in instance.customParameters)

    assert instance.customParameters[0].value == "ccmp;sub space by space;"
    assert instance.customParameters[1].value == "liga;sub space space by space;"


@pytest.mark.parametrize("disabled", [False, True])
def test_disabled_glyphOrder_custom_params(ufo_module, disabled):
    # With minimal=True, 'disabled' custom parameters should be ignored
    # https://github.com/googlefonts/glyphsLib/issues/905
    # https://github.com/googlefonts/fontc/issues/985
    font = GSFont()
    font.masters.append(GSFontMaster())

    implicit_glyph_order = [".notdef", "A", "B", "C"]
    for glyph_name in implicit_glyph_order:
        font.glyphs.append(GSGlyph(glyph_name))

    custom_glyph_order = [".notdef", "C", "B", "A"]
    font.customParameters.append(
        GSCustomParameter("glyphOrder", custom_glyph_order, disabled=disabled)
    )

    ufo = to_ufos(font, ufo_module=ufo_module, minimal=True)[0]

    if disabled:
        assert ufo.lib["public.glyphOrder"] == implicit_glyph_order
    else:
        assert ufo.lib["public.glyphOrder"] == custom_glyph_order
